from __future__ import annotations

import json
import time
from configparser import ConfigParser
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..persistent.parameters import ParameterValue
from ..persistent.parameters import _format_info_category
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils import directories
from .flint_tomo_imshow import SingleSliceImshow
from .tomo_model import TomoProcessorModel
from .utils import calculate_relative_CoR_estimate
from .utils import get_estimate_cor_metadata

SUPPORTED_TOMO_SCANS = ["tomo:basic", "tomo:zseries", "tomo:fullturn", "tomo:halfturn"]
# The following scan types are not yet supported:
# tomo:helical, tomo:zhelical, tomo:ptychotomo, tomo:multitomo , tomo:holotomo


class TomoProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo(
            "slice_reconstruction_workflow",
            category="workflows",
            doc="Workflow file for slice reconstruction",
        ),
        ParameterInfo(
            "volume_reconstruction_workflow",
            category="workflows",
            doc="Workflow file for volume reconstruction",
        ),
        ParameterInfo("queue", category="workflows"),
        ParameterInfo(
            "_bliss_hdf5_path",
            category="files",
            doc="HDF5 Dataset path (filled automatically)",
        ),
        ParameterInfo(
            "_output_path",
            category="files",
            doc="Nx output path (filled automatically)",
        ),
        ParameterInfo(
            "nabu_config_file",
            category="slice_reconstruction_parameters",
            doc="Nabu configuration file",
        ),
        ParameterInfo(
            "_reference_dir_to_soft_link",
            category="files",
            doc="Directory from which the reduced darks and flats will be linked.",
        ),
        ParameterInfo(
            "_flip_left_right",
            category="estimate_center_of_rotation",
            doc="If True, the center of rotation estimate is flipped",
        ),
        ParameterInfo(
            "offset_mm",
            category="estimate_center_of_rotation",
            doc="Offset (mm) subtracted from the translation_y motor position",
        ),
        ParameterInfo(
            "_orientation_factor",
            category="estimate_center_of_rotation",
            doc="Orientation factor to be multiplied to the translation_y motor value",
        ),
        ParameterInfo(
            "_mechanical_ud_flip",
            category="slice_reconstruction_parameters",
            doc="If True, adds a NXtransformation in the NX file to describe a flip up-down",
        ),
        ParameterInfo(
            "_mechanical_lr_flip",
            category="slice_reconstruction_parameters",
            doc="If True, adds a NXtransformation in the NX file to describe a flip left-right",
        ),
        ParameterInfo(
            "cor_algorithm",
            category="estimate_center_of_rotation",
            doc="Method to estimate the centre of rotation in the frame",
        ),
        ParameterInfo(
            "estimated_cor",
            category="estimate_center_of_rotation",
            doc="Pixel value of the CoR in the frame (relative, filled automatically)",
        ),
        ParameterInfo(
            "slice_index",
            category="slice_reconstruction_parameters",
            doc="Index of the slice that will be reconstructed online",
        ),
        ParameterInfo(
            "phase_retrieval_method",
            category="slice_reconstruction_parameters",
            doc="Phase retrieval method or 'None'",
        ),
        ParameterInfo(
            "delta_beta",
            category="slice_reconstruction_parameters",
            doc="For Paganin or CTF phase retrieval, default is 100",
        ),
        ParameterInfo(
            "volume_reconstruction",
            category="volume_reconstruction_parameters",
            doc="If True, performs volume reconstruction after slice reconstruction",
        ),
        ParameterInfo(
            "slice_index_range",
            category="volume_reconstruction_parameters",
            doc="Range of slice indices to reconstruct the volume, (z_bottom, z_top) or 'all'",
        ),
        ParameterInfo(
            "output_format",
            category="volume_reconstruction_parameters",
            doc="Output format for the volume reconstruction: 'hdf5', 'tiff', 'jp2', 'edf' or 'vol'",
        ),
        ParameterInfo(
            "show_last_slice",
            category="flint_display_parameters",
            doc="If True, displays the last reconstructed slice in Flint",
        ),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Initialize the TomoProcessor for converting BLISS HDF5 data to Nexus (NX) format.

        This processor can be integrated into a BLISS beamline configuration. Typical
        usage involves adding the following to a configuration yaml file:

        - name: tomo_blissoda
          plugin: generic
          class: TomoProcessor
          package: blissoda.tomo.tomo_processor
        """

        if defaults is None:
            defaults = {}
        defaults.setdefault("trigger_at", "END")
        defaults.setdefault(
            "slice_reconstruction_workflow", "slice_reconstruction.json"
        )
        defaults.setdefault(
            "volume_reconstruction_workflow", "volume_reconstruction.json"
        )
        defaults.setdefault("nabu_config_file", None)
        defaults.setdefault("_reference_dir_to_soft_link", None)
        defaults.setdefault("slice_index", "middle")
        defaults.setdefault("cor_algorithm", "sliding-window")
        defaults.setdefault("phase_retrieval_method", "None")
        defaults.setdefault("delta_beta", "100")
        defaults.setdefault("_flip_left_right", False)
        defaults.setdefault("offset_mm", 0.0)
        defaults.setdefault("_orientation_factor", 1.0)
        defaults.setdefault("_mechanical_ud_flip", False)
        defaults.setdefault("_mechanical_lr_flip", False)
        defaults.setdefault("estimated_cor", 0.0)
        defaults.setdefault("volume_reconstruction", False)
        defaults.setdefault("slice_index_range", "all")
        defaults.setdefault("output_format", "tiff")
        defaults.setdefault("show_last_slice", False)
        super().__setattr__("_tomo_model", TomoProcessorModel(**defaults))

        self.imshow = SingleSliceImshow(history=1)
        super().__init__(config=config, defaults=defaults)

    def __setattr__(self, name, value):
        if hasattr(self, "_tomo_model") and hasattr(self._tomo_model, name):
            setattr(self._tomo_model, name, value)
            value = getattr(self._tomo_model, name)

        super().__setattr__(name, value)

    def __info__(self) -> str:
        self.estimated_cor = self.estimate_CoR()
        categories = self._info_categories()
        for category in categories.values():
            for key, info in list(category.items()):
                if key in dir(self) and isinstance(info, ParameterValue):
                    category[key] = ParameterValue(getattr(self, key), info.doc)

        # Does not display delta_beta if phase_retrieval_method is None
        parameters = categories.get("slice_reconstruction_parameters")
        if parameters is not None:
            method = parameters.get("phase_retrieval_method")
            if method is not None and method.value == "None":
                parameters.pop("delta_beta", None)

        # Does not display volume reconstruction parameters if volume_reconstruction is False
        volume_params = categories.get("volume_reconstruction_parameters")
        if volume_params is not None:
            volume_reconstruction = volume_params.get("volume_reconstruction")
            if volume_reconstruction is not None and not volume_reconstruction.value:
                volume_params.pop("slice_index_range", None)
                volume_params.pop("output_format", None)

        return "\n" + "\n\n".join(
            [
                f"{name.replace('_', ' ').title()}:\n {_format_info_category(category)}"
                for name, category in categories.items()
                if category
            ]
        )

    def load_nabu_config_file(self, nabu_file, allow_no_value=False) -> Dict[str, Any]:
        """
        Parse a configuration file and returns a dictionary.
        """
        if nabu_file and Path(nabu_file).exists():
            parser = ConfigParser(
                inline_comment_prefixes=("#",),
                allow_no_value=allow_no_value,
            )
            with open(nabu_file) as fid:
                file_content = fid.read()
            parser.read_string(file_content)
            nabu_dict = parser._sections
        else:
            nabu_dict = dict()

        # Ensure nested dictionaries exist
        nabu_dict.setdefault("dataset", {})
        nabu_dict.setdefault("reconstruction", {})
        darks_flats_dir = Path(self._output_path).parent.parent / "references"
        nabu_dict["dataset"]["darks_flats_dir"] = str(darks_flats_dir)
        nabu_dict["reconstruction"]["rotation_axis_position"] = self.cor_algorithm
        current_cor_options = nabu_dict["reconstruction"].get("cor_options")
        nabu_dict["reconstruction"]["cor_options"] = self._update_cor_options(
            current_cor_options, self.estimated_cor
        )
        nabu_dict.setdefault("phase", {})
        nabu_dict["phase"]["method"] = self.phase_retrieval_method
        nabu_dict["phase"]["delta_beta"] = self.delta_beta
        nabu_dict.setdefault("output", {})

        return nabu_dict

    def _update_cor_options(
        self, existing_options: Optional[str], side_value: Any
    ) -> str:
        """Replace the 'side' option and keep any others as-is."""
        if not existing_options:
            return f"side={side_value}"

        options = [opt.strip() for opt in existing_options.split(";") if opt.strip()]
        options = [opt for opt in options if not opt.startswith("side")]
        return "; ".join([f"side={side_value}", *options])

    def estimate_CoR(self) -> float:
        """
        Estimate the relative center of rotation based on scan metadata and processor parameters.
        """

        pixel_size_mm, _, translation_y_mm = get_estimate_cor_metadata()

        center_of_rotation = calculate_relative_CoR_estimate(
            pixel_size_mm=pixel_size_mm,
            translation_y_mm=translation_y_mm,
            offset_mm=self.offset_mm,
            flip=self._flip_left_right,
            orientation_factor=self._orientation_factor,
        )
        return center_of_rotation

    def _get_scan_parameters(self, scan: BlissScanType) -> Dict[str, Any]:
        scan_parameters = dict()
        for key in list(scan.scan_info.keys()):
            scan_parameters[key] = scan.scan_info[key]
        return scan_parameters

    def _build_output_path(self, bliss_path: str) -> str:
        """
        Build the Nx output path under PROCESSED_DATA/sample/sample_dataset/projections.
        """
        processed_path = bliss_path.replace("RAW_DATA", "PROCESSED_DATA")
        nx_path = Path(processed_path).with_suffix(".nx")
        projections_dir = nx_path.parent / "projections"
        return str(projections_dir / nx_path.name)

    def _build_darks_flats_dir_path(self, _output_path: str, scan) -> None:
        """
        Build the references output path under PROCESSED_DATA/sample/sample_dataset/references if
        the scan contains the required dark and flat images.
        """
        if self._has_darks_and_flats(scan):
            darks_flats_dir = Path(_output_path).parent.parent / "references"
            # This is kept and only used for the ReduceDarkFlat that will create soft link
            self._reference_dir_to_soft_link = str(darks_flats_dir)
        elif self._reference_dir_to_soft_link is None:
            raise ValueError("Darks and flats directory cannot be determined")

    def _has_darks_and_flats(self, scan: BlissScanType) -> bool:
        """
        Check if the scan contains at least one dark and one flat, regardless of their position.
        """
        scan_parameters = self._get_scan_parameters(scan)
        scan_flags = scan_parameters.get("technique", {}).get("scan_flags", {})

        has_dark_images = any(
            bool(scan_flags.get(key, False))
            for key in ("dark_images_at_start", "dark_images_at_end")
        )
        has_flat_images = any(
            bool(scan_flags.get(key, False))
            for key in ("ref_images_at_start", "ref_images_at_end")
        )

        return has_dark_images and has_flat_images

    def _get_inputs(self, scan: BlissScanType) -> List[Dict[str, Any]]:
        scan_parameters = self._get_scan_parameters(scan)
        self.estimated_cor = self.estimate_CoR()
        self._bliss_hdf5_path = scan_parameters["filename"]
        self._output_path = self._build_output_path(self._bliss_hdf5_path)
        self._build_darks_flats_dir_path(self._output_path, scan)
        slice_index_range = (
            None if self.slice_index_range == "all" else self.slice_index_range
        )

        inputs = list()
        inputs.append(
            {
                "task_identifier": "ewokstomo.tasks.nxtomomill.H5ToNx",
                "name": "bliss_hdf5_path",
                "value": self._bliss_hdf5_path,
            }
        )
        inputs.append(
            {
                "task_identifier": "ewokstomo.tasks.nxtomomill.H5ToNx",
                "name": "nx_path",
                "value": self._output_path,
            }
        )
        inputs.append(
            {
                "task_identifier": "ewokstomo.tasks.nxtomomill.H5ToNx",
                "name": "mechanical_ud_flip",
                "value": self._mechanical_ud_flip,
            }
        )
        inputs.append(
            {
                "task_identifier": "ewokstomo.tasks.nxtomomill.H5ToNx",
                "name": "mechanical_lr_flip",
                "value": self._mechanical_lr_flip,
            }
        )

        if not self._has_darks_and_flats(scan):
            inputs.append(
                {
                    "task_identifier": "ewokstomo.tasks.reducedarkflat.ReduceDarkFlat",
                    "name": "reference_dir_to_soft_link",
                    "value": self._reference_dir_to_soft_link,
                }
            )

        inputs.append(
            {
                "task_identifier": "ewokstomo.tasks.reconstruct_slice.ReconstructSlice",
                "name": "nx_path",
                "value": self._output_path,
            }
        )
        nabu_dict = self.load_nabu_config_file(self.nabu_config_file)
        inputs.append(
            {
                "task_identifier": "ewokstomo.tasks.reconstruct_slice.ReconstructSlice",
                "name": "config_dict",
                "value": nabu_dict,
            }
        )
        inputs.append(
            {
                "task_identifier": "ewokstomo.tasks.reconstruct_slice.ReconstructSlice",
                "name": "slice_index",
                "value": self.slice_index,
            }
        )
        if self.volume_reconstruction:
            inputs.append(
                {
                    "task_identifier": "ewokstomo.tasks.reconstruct_volume.ReconstructVolume",
                    "name": "nx_path",
                    "value": self._output_path,
                }
            )
            inputs.append(
                {
                    "task_identifier": "ewokstomo.tasks.reconstruct_volume.ReconstructVolume",
                    "name": "slice_index_range",
                    "value": slice_index_range,
                }
            )
            volume_nabu_dict = self.load_nabu_config_file(self.nabu_config_file)
            volume_nabu_dict["output"]["file_format"] = "hdf5"
            inputs.append(
                {
                    "task_identifier": "ewokstomo.tasks.reconstruct_volume.ReconstructVolume",
                    "name": "config_dict",
                    "value": volume_nabu_dict,
                }
            )
            inputs.append(
                {
                    "task_identifier": "ewokstomo.tasks.convert_volume.ConvertVolumeTo16Bit",
                    "name": "output_format",
                    "value": self.output_format,
                }
            )
            inputs.append(
                {
                    "id": "dataportal_task_volumes",
                    "task_identifier": "ewokstomo.tasks.dataportalupload.DataPortalUpload",
                    "name": "dataset",
                    "value": "volumes",
                }
            )
        inputs.append(
            {
                "id": "dataportal_task_projections",
                "task_identifier": "ewokstomo.tasks.dataportalupload.DataPortalUpload",
                "name": "dataset",
                "value": "projections",
            }
        )
        inputs.append(
            {
                "id": "dataportal_task_slices",
                "task_identifier": "ewokstomo.tasks.dataportalupload.DataPortalUpload",
                "name": "dataset",
                "value": "slices",
            }
        )
        return inputs

    def _get_workflows_dir(self, dataset_filename: str) -> Path:
        processed_dir = Path(directories.get_dataset_processed_dir(dataset_filename))
        return processed_dir / "workflows" / "gallery"

    def _get_workflow_upload_parameters(
        self, scan: BlissScanType
    ) -> Optional[Dict[str, Any]]:
        if not scan.scan_info.get("save"):
            return None
        scan_saving = current_session.scan_saving
        filename = scan.scan_info.get("filename") or scan_saving.filename
        metadata = {"Sample_name": scan_saving.dataset["Sample_name"]}
        workflows_dir = self._get_workflows_dir(filename)
        raw_directory = str(Path(filename).parent)
        return {
            "beamline": scan_saving.beamline,
            "proposal": scan_saving.proposal_name,
            "dataset": "workflows",
            "path": str(workflows_dir),
            "raw": [raw_directory],
            "metadata": metadata,
        }

    def _get_workflow(self):
        if self.volume_reconstruction:
            workflow = self.volume_reconstruction_workflow
        else:
            workflow = self.slice_reconstruction_workflow
        with open(resource_filename("tomo", workflow), "r") as wf:
            return json.load(wf)

    def _get_submit_arguments(self, scan) -> Dict[str, Any]:
        inputs = self._get_inputs(scan)
        kwargs = {
            "inputs": inputs,
            "outputs": [{"all": False}],
            "engine": "dask",
            "scheduler": "multithreading",
        }
        upload_parameters = self._get_workflow_upload_parameters(scan)
        if upload_parameters:
            kwargs["upload_parameters"] = upload_parameters
        return kwargs

    def workflow_destination(self) -> str:
        """
        Returns the destination path for the workflow output.
        """
        workflows_dir = self._get_workflows_dir(self._bliss_hdf5_path)
        filename = Path(self._output_path).with_suffix(".json").name
        return str(workflows_dir / filename)

    def execute_workflow(self, scan: BlissScanType) -> None:
        if (
            "tomoconfig" not in scan.scan_info.get("technique", "")
            or not scan.scan_info["title"] in SUPPORTED_TOMO_SCANS
        ):
            return
        workflow = self._get_workflow()
        kwargs = self._get_submit_arguments(scan)
        kwargs["convert_destination"] = self.workflow_destination()
        time.sleep(2)
        submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
        if self.show_last_slice:
            self.imshow._spawn(self.imshow.monitor_and_display_slice, self._output_path)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> None:
        self.execute_workflow(scan)



from ...bliss_globals import setup_globals
from .. import testing
from ..processors.id32 import DemoId32pecGenPProcessor


@testing.integration_fixture
def id32_specgen_processor():
    processor = DemoId32pecGenPProcessor()
    processor.enable()
    yield processor
    processor.disable()


def id32_specgen_demo(expo=0.2, npoints=10):
    itest_id32(expo=expo, npoints=npoints)


@testing.integration_test
def itest_id32(id32_specgen_processor, expo=0.2, npoints=10):
    print("The workflow is expected to fail so this only tests the triggering for now.")
    setup_globals.loopscan(npoints, expo, setup_globals.difflab6)
    testing.wait_workflows()



from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ...bliss_globals import setup_globals
from ...id09.txs_processor import TxsProcessor


class DemoTxsProcessor(TxsProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)
        defaults.setdefault("queue", "celery")
        defaults.setdefault("detector", "difflab6")
        defaults.setdefault("pixel", (10e-3, 10e-3))
        defaults.setdefault("energy", 10e3)

        super().__init__(config=config, defaults=defaults)

    def pct(self, *args, **kw):
        s = setup_globals.ct(*args, **kw)
        self.trigger_workflow_on_new_scan(s)
        return s

    @property
    def lima_names(self) -> List[str]:
        return [self.detector]

    def _expected_data_shapes_1d(self, lima_name: str, npoints: int):
        return {
            f"{lima_name}:q": (600,),
            f"{lima_name}:intensity": (npoints, 600),
            f"{lima_name}:points": (npoints,),
        }



import numpy

from ...bliss_globals import setup_globals
from ...version_utils import has_minimal_version
from .. import testing
from ..processors.stop_scan import DemoPreset


@testing.integration_test
def stop_scan_demo():
    if not has_minimal_version("bliss", "2.2"):
        testing.skip_integration_test("Requires bliss>=2.2")

    expo = 0.2
    detectors = [setup_globals.diode1, setup_globals.difflab6]

    # Stop criterium is based the total sum of all data of several detectors.
    # Get a reasonable limit for the total.
    scan = setup_globals.ct(expo, *detectors)
    max_difflab6 = numpy.nanmax(scan.streams["difflab6:image"][:])
    max_diode1 = numpy.nanmax(
        scan.streams["simulation_diode_sampling_controller:diode1"][:]
    )
    lima_limits = {"difflab6": float(max_difflab6 * 5)}
    scalar_counter_limits = {"diode1": float(max_diode1 * 5)}

    # Scan with preset to start/stop data processing and handle feedback streamed through Redis.
    p = DemoPreset(lima_limits, scalar_counter_limits)
    scan = setup_globals.loopscan(100, expo, *detectors, run=False)
    scan.acq_chain.add_preset(p)
    scan.run()

    # Check that the scan stopped early
    for name in ("difflab6:image", "simulation_diode_sampling_controller:diode1"):
        npoints = len(scan.streams[name])
        assert (
            npoints < 20
        ), f"Stream {name:!r} has {npoints} which is more than expected"



from __future__ import annotations

import functools
import logging
import time
from collections import OrderedDict
from pathlib import Path

import h5py
import numpy as np

from blissoda.tomo.utils import apply_labels
from blissoda.tomo.utils import compute_axes

from ..bliss_globals import current_session
from ..flint.plotter import BasePlotter

try:
    from flint.viewers.custom_image.client import ImageView
except ImportError as ex:
    from ..import_utils import unavailable_class

    ImageView = unavailable_class(ex)

logger = logging.getLogger(__name__)


def _background_task(method):
    """Decorator to handle exceptions in background tasks."""

    @functools.wraps(method)
    def wrapper(*args, **kw):
        try:
            return method(*args, **kw)
        except Exception as e:
            logger.error("Online tomo plotter task failed (%s)", e, exc_info=True)
            raise

    return wrapper


class OnlineTomoAccumulatedPlotter(BasePlotter):
    """
    Plotter for online tomography reconstruction that accumulates partial slices.

    Monitors the output directory for new reconstructed slice files and
    progressively accumulates them to show the evolving reconstruction.
    """

    TITLE = "Accumulated Reconstructed Slice"

    def __init__(self, history: int = 1, retry_period: int = 3) -> None:
        """
        Initialize the plotter.

        :param history: Number of plots to keep (should be 1 for current scan only).
        :param retry_period: Polling interval in seconds to check for new batch files.

        """
        super().__init__(max_plots=history)
        self.retry_period = retry_period
        self._cache: OrderedDict[str, np.ndarray] = OrderedDict()
        self._current_plot_widget = None
        self._accumulated_slice = None
        self._batch_count = 0
        self._batch_size = None

    def handle_workflow_result(
        self,
        future,
        output_path: str,
        slice_index: int | str = "middle",
        batch_size: int = 100,
    ) -> None:
        """
        Monitor workflow execution and update plot as new batches are saved.

        :param future: The workflow job future
        :param output_path: Directory where batch files are being saved
        :param slice_index: The slice index being reconstructed (for display only)
        :param batch_size: Number of projections per batch

        """

        self._spawn(
            self._monitor_reconstruction,
            future,
            output_path,
            slice_index,
            batch_size,
        )

    @_background_task
    def _monitor_reconstruction(
        self,
        future,
        output_path: str,
        slice_index: int | str,
        batch_size: int,
    ) -> None:
        """
        Background task to monitor reconstruction progress.

        Polls the output directory for new batch files and updates the plot.
        :param future: The workflow job future
        :param output_path: Directory where batch files are being saved
        :param slice_index: The slice index being reconstructed (for display only)
        :param batch_size: Number of projections per batch
        """
        output_dir = Path(output_path)

        # Reset state for new reconstruction
        self._accumulated_slice = None
        self._batch_count = 0
        self._batch_size = batch_size

        logger.info(f"Starting monitoring of reconstruction in {output_dir}")

        last_processed_count = 0

        # Monitor while workflow is running
        while not future.done():
            last_processed_count = self._poll_for_batches(
                output_dir, last_processed_count
            )
            if last_processed_count != self._batch_count:
                self._update_plot(slice_index)
            time.sleep(self.retry_period)

        # Final update after workflow completes
        try:
            self._finalize_monitoring(output_dir, last_processed_count, slice_index)
        except Exception as e:
            logger.error(f"Error in final batch processing: {e}")

    def _poll_for_batches(self, output_dir: Path, start_count: int) -> int:
        """
        Poll the output directory for new batch files and process them.
        :param output_dir: Directory containing the batch slice files
        :param start_count: Number of batches already processed
        :return: Updated number of batches processed
        """
        try:
            new_batches = self._process_new_batches(output_dir, start_count)
            if new_batches > 0:
                return self._batch_count
        except Exception as e:
            logger.debug(f"Error processing batches: {e}")
        return start_count

    def _finalize_monitoring(
        self, output_dir: Path, last_processed_count: int, slice_index
    ):
        """
        Final processing after workflow completion.
        :param output_dir: Directory containing the batch slice files
        :param last_processed_count: Number of batches already processed
        :param slice_index: The slice index being reconstructed
        """
        self._process_new_batches(output_dir, last_processed_count)
        self._update_plot(slice_index)

        projections_done = self._batch_count * self._batch_size
        logger.info(
            f"Reconstruction monitoring completed: "
            f"{projections_done} projections processed."
        )

    def _process_new_batches(self, output_dir: Path, start_count: int) -> int:
        """
        Process any new batch slice files that have appeared.

        :param output_dir: Directory containing the batch slice files
        :param start_count: Number of batches already processed

        :return: Number of new batches processed

        """
        if not output_dir.exists():
            return 0

        # Find all batch files
        batch_files = sorted(output_dir.glob("*.h5"))
        if not batch_files:
            return 0

        new_batches = 0

        # Process only new files
        for batch_file in batch_files[start_count:]:
            try:
                with h5py.File(batch_file, "r") as f:
                    if "reconstructed_slice" not in f:
                        logger.debug(f"Missing dataset in {batch_file}")
                        continue

                    slice_data = f["reconstructed_slice"][()]

                    # Accumulate by addition
                    if self._accumulated_slice is None:
                        self._accumulated_slice = slice_data.astype(np.float32)
                    else:
                        self._accumulated_slice += slice_data

                    self._batch_count += 1
                    new_batches += 1
                    logger.debug(f"Processed batch file: {batch_file.name}")

            except (OSError, KeyError) as e:
                logger.debug(f"Could not read {batch_file}: {e}")
                continue

        return new_batches

    def _update_plot(self, slice_index: int | str) -> None:
        """
        Update the Flint plot with the accumulated slice.

        :param slice_index: The slice index being reconstructed
        """

        if self._accumulated_slice is None:
            return

        widget = self._get_plot(self.TITLE, ImageView)

        # Set title with progress information
        self._set_title_with_progress(widget, slice_index)

        # Compute physical axes from tomo config
        x_axis, y_axis = compute_axes(self._accumulated_slice)
        origin = (float(x_axis[0]), float(y_axis[0]))

        dx = float(x_axis[1] - x_axis[0]) if len(x_axis) > 1 else 1.0
        dy = float(y_axis[1] - y_axis[0]) if len(y_axis) > 1 else 1.0
        scale = (dx, dy)

        # Update the image data
        widget.set_data(self._accumulated_slice, origin=origin, scale=scale)

        # Apply axis labels
        apply_labels(widget)

        # Cache the current image
        self._cache[self.TITLE] = self._accumulated_slice

        self.purge_tasks()
        self._purge()

    def _set_title_with_progress(
        self, widget: ImageView, slice_index: int | str
    ) -> None:
        """
        Set the plot window title with progress information.

        :param widget: The plot widget
        :param slice_index: The slice index being reconstructed
        """
        base_title = current_session.scan_saving.data_filename

        # Add progress and slice info to title
        nb_projections = self._batch_count * self._batch_size
        title = f"{base_title} - Slice {slice_index} ({nb_projections} projections processed)"
        widget.title = title

    def clear(self) -> None:
        """Clear the current plot and reset state."""
        self._accumulated_slice = None
        self._batch_count = 0
        logger.debug("Plotter state cleared")

    def _purge(self) -> None:
        """
        Remove oldest images beyond history.
        """
        while len(self._cache) > self._max_plots:
            self._cache.popitem(last=False)



"""Integration tests for the id11 XRPD and PDF processing."""

from ...bliss_globals import setup_globals
from .. import testing
from ..processors.id11 import DemoId11XrpdProcessor
from .utils import xrpd


@testing.integration_fixture
def id11_processor():
    xrpd_processor = DemoId11XrpdProcessor()
    xrpd_processor.enable(setup_globals.difflab6)
    xrpd_processor._plotter.clear_lima_plots("difflab6")

    xrpd_processor.pdf_enable = False
    # TODO:
    # 1. Only enable when diffpy (proprietary) is installed in the worker environment.
    # 2. PdfGetXSaveNexus has missing parameters 'output_filename', 'detector_name', 'scan'.
    # 3. Validate PDF results in addition to XRPD results.

    yield xrpd_processor

    xrpd_processor.disable()


def id11_demo(expo=0.2, npoints=10):
    test_id11_ct_with_1d_integration(expo=expo)
    test_id11_scan_with_1d_integration(expo=expo, npoints=npoints)


@testing.integration_test
def test_id11_ct_with_1d_integration(
    debug_flint, id11_processor, nrepeats: int = None, expo=0.2
):
    xrpd.test_xrpd_ct_with_1d_integration(
        id11_processor,
        [setup_globals.difflab6, setup_globals.diode1, setup_globals.diode2],
        nrepeats=nrepeats,
        expo=expo,
    )


@testing.integration_test
def test_id11_scan_with_1d_integration(
    debug_flint, id11_processor, nrepeats: int = None, expo=0.2, npoints=10
):
    xrpd.test_xrpd_scan_with_1d_integration(
        [id11_processor],
        [
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        ],
        nrepeats=nrepeats,
        expo=expo,
        npoints=npoints,
    )



from .. import testing
from ..processors.streamline import DemoStreamlineScanner
from .utils import streamline


def streamline_demo():
    test_streamline_without_autocalibration()
    test_streamline_with_autocalibration()


@testing.integration_fixture
def streamline_scanner():
    return DemoStreamlineScanner()


@testing.integration_test
def test_streamline_with_autocalibration(streamline_scanner):
    streamline.test_streamline_with_autocalibration(streamline_scanner)


@testing.integration_test
def test_streamline_without_autocalibration(streamline_scanner):
    streamline.test_streamline_without_autocalibration(streamline_scanner)



from blissoda.demo.tests.itest_xrpd import DemoXrpdProcessor  # noqa F401
from blissoda.demo.tests.itest_xrpd import xrpd_demo_1d  # noqa F401
from blissoda.demo.tests.itest_xrpd import xrpd_demo_2d  # noqa F401



import pathlib

from ...bliss_globals import setup_globals
from .. import testing
from ..processors.fluo import DemoFluoProcessor


@testing.integration_fixture
def fluo_processor():
    processor = DemoFluoProcessor()
    processor.enable()
    yield processor
    processor.disable()


def fluo_demo(expo=0.2, shape=(3, 4)):
    testfluo_xrfmap(expo=expo, shape=shape)


@testing.integration_test
def testfluo_xrfmap(fluo_processor, expo=0.2, shape=(3, 4)):
    robx_start = 0
    robx_step_size = 0.01
    robx_intervals = shape[1] - 1
    robx_end = robx_start + robx_step_size * robx_intervals

    roby_start = 0
    roby_step_size = -0.01
    roby_intervals = shape[0] - 1
    roby_end = roby_start + roby_step_size * roby_intervals

    _ = setup_globals.amesh(
        setup_globals.robx,
        robx_start,
        robx_end,
        robx_intervals,
        setup_globals.roby,
        roby_start,
        roby_end,
        roby_intervals,
        expo,
        setup_globals.mca1,
        setup_globals.mca2,
        setup_globals.diode1,
    )

    assert fluo_processor._last_future is not None, "Fluo workflow not triggered"
    result = fluo_processor._last_future.result(timeout=20)

    for key in ("output_root_uri", "xrf_results_uri"):
        file_path, data_path = result[key].split("::")
        testing.assert_hdf5_group_exists(pathlib.Path(file_path), data_path)

    file_path, data_path = result["xrf_results_uri"].split("::")
    data_path += "/parameters/Cr_K"
    testing.assert_hdf5_dataset_exists(pathlib.Path(file_path), data_path, shape)



from blissoda.demo.tests.itest_bm08 import DemoBm08Hdf5ToXdiConverter  # noqa F401
from blissoda.demo.tests.itest_bm08 import bm08_demo  # noqa F401
from blissoda.demo.tests.itest_exafs_bm08 import DemoBm08ExafsProcessor  # noqa F401
from blissoda.demo.tests.itest_exafs_bm08 import bm08_exafs_demo  # noqa F401



from blissoda.demo.processors.id22 import DemoStScanProcessor  # noqa F401
from blissoda.demo.processors.xrpd_id22 import DemoId22XrpdProcessor  # noqa F401
from blissoda.demo.tests.itest_id22 import id22_stscan_demo  # noqa F401
from blissoda.demo.tests.itest_xrpd_id22 import id22_xrpd_demo  # noqa F401



from ....bliss_globals import setup_globals
from .pyfai_calib import DEFAULT_CALIB


def lima_image_is_padded(lima_name: str) -> bool:
    if DEFAULT_CALIB[lima_name]["detector"] == "Pilatus1M":
        # lima-camera-simulator<1.9.10 does not support odd image widths
        # so Pilatus1M data is padded with one pixel to make it even.
        lima = getattr(setup_globals, lima_name)
        return not bool(lima.image.width % 2)
    return False



from typing import Any
from typing import Dict
from typing import Optional

from ...bm08.converter import Bm08Hdf5ToXdiConverter
from ...processor import BlissScanType


class DemoBm08Hdf5ToXdiConverter(Bm08Hdf5ToXdiConverter):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = dict()
        defaults.setdefault("mono_counter", "roby")  # Must have no units
        defaults.setdefault("crystal_motor", "sy")
        defaults.setdefault("optional_counters", ["diode1", "diode2"])
        defaults.setdefault("optional_mca_counters", ["OdaRoi"])
        defaults.setdefault("retry_timeout", 15)
        defaults.setdefault("queue", "celery")
        super().__init__(config=config, defaults=defaults)

    def _get_workflow_upload_parameters(self, scan: BlissScanType) -> None:
        return None



import os

from ...streamline.scanner import StreamlineScanner
from .utils.streamline import DemoStreamlineScannerMixIn


class DemoStreamlineScanner(DemoStreamlineScannerMixIn, StreamlineScanner):
    _TEST_DATA_PATHS_WITHOUT_CALIB = (
        "/results/integrate/integrated/q",
        "/results/difflab6_integrate_2th/integrated/2th",
    )
    _TEST_DATA_PATHS_WITH_CALIB = (
        "/results/integrate/integrated/q",
        "/results/difflab6_integrate_2th/integrated/2th",
    )
    _TEST_ASCII_SUFFIXES_WITHOUT_CALIB = tuple()
    _TEST_ASCII_SUFFIXES_WITH_CALIB = tuple()

    def calib(self, *args, **kwargs):
        self._requires_workflow()
        super().calib(*args, **kwargs)

    def run(self, *args, **kwargs):
        self._requires_workflow()
        super().run(*args, **kwargs)

    def _requires_workflow(self):
        if not os.path.exists(self.workflow):
            raise RuntimeError(
                "the workflow file no longer exists, execute 'init_workflow' again"
            )



from blissoda.demo.tests.itest_id01 import DemoCdiProcessor  # noqa F401
from blissoda.demo.tests.itest_id01 import id01_cdi_demo  # noqa F401
from blissoda.demo.tests.itest_id01 import scan_sequence  # noqa F401



import os
from collections import namedtuple
from contextlib import ExitStack

import mock
import pytest

from ..persistent import mock_redis
from ..persistent import ordereddict as ordereddict_mod
from ..persistent import parameters as parameters_mod
from ..version_utils import has_minimal_version


def has_redis_server() -> bool:
    # TODO: instead of pytest-redis which needs a real Redis server, use fakeredis?
    with os.popen("redis-server --version") as output:
        return bool(output.read())


MockedBlissInfo = namedtuple("MockedBlissInfo", "beacon_connection")

if has_redis_server():

    @pytest.fixture(scope="session")
    def mock_bliss_session(redis_proc):
        # Ensure we do not connect to a real Beacon host
        os.environ["BEACON_HOST"] = "nonexisting:25000"

        # Skip test when Beacon dependencies are not available
        # TODO: mock to the extend we no longer need bliss as a dependency
        if has_minimal_version("blissdata", "1"):
            try:
                from bliss.redis.manager import RedisAddress
            except ImportError:
                pytest.skip("requires bliss")
        else:
            from blissdata.redis import RedisAddress

        try:
            from bliss.config.conductor.client import get_default_connection
            from bliss.config.conductor.connection import Connection
        except ImportError:
            pytest.skip("requires bliss")

        redis_url = f"{redis_proc.host}:{redis_proc.port}"

        def _get_redis_url(self):
            """Return the pytest-redis server URL"""
            return RedisAddress.factory(redis_url)

        with ExitStack() as stack:
            # The Beacon connection should connect to the pytest-redis server
            ctx = mock.patch.object(
                Connection, "get_redis_connection_address", _get_redis_url
            )
            stack.enter_context(ctx)
            ctx = mock.patch.object(
                Connection, "get_redis_data_server_connection_address", _get_redis_url
            )
            stack.enter_context(ctx)

            yield MockedBlissInfo(get_default_connection())

else:

    @pytest.fixture(scope="session")
    def mock_bliss_session():
        # redis_proc raises an exception when redis-server is not available
        pytest.skip("requires redis-server")


@pytest.fixture
def mock_bliss(mock_bliss_session):
    yield
    proxy = mock_bliss_session.beacon_connection.get_redis_proxy()
    proxy.flushall()


@pytest.fixture
def mock_persistent():
    with mock.patch.object(
        parameters_mod, "HashObjSetting", mock_redis.LocalHashObjSetting
    ):
        with mock.patch.object(
            ordereddict_mod,
            "OrderedHashObjSetting",
            mock_redis.LocalOrderedHashObjSetting,
        ):
            try:
                mock_redis.LocalHashObjSetting._LOCAL_REASON = "unit test"
                mock_redis.LocalOrderedHashObjSetting._LOCAL_REASON = "unit test"
                mock_redis.LocalHashObjSetting._LOCAL_WARNING_EMITTED = False
                mock_redis.LocalOrderedHashObjSetting._LOCAL_WARNING_EMITTED = False

                yield mock_redis._REDIS_STORE
            finally:
                mock_redis._REDIS_STORE.clear()

                mock_redis.LocalHashObjSetting._LOCAL_REASON = None
                mock_redis.LocalOrderedHashObjSetting._LOCAL_REASON = None
                mock_redis.LocalHashObjSetting._LOCAL_WARNING_EMITTED = False
                mock_redis.LocalOrderedHashObjSetting._LOCAL_WARNING_EMITTED = False


@pytest.fixture
def mock_processing_triggers():
    enabled = {}

    def enable(processing_id, trigger, timing):
        enabled[processing_id] = (trigger, timing)

    def disable(processing_id):
        enabled.pop(processing_id, None)

    with mock.patch(
        "blissoda.utils.trigger.enable_processing_trigger",
        side_effect=enable,
        autospec=True,
    ), mock.patch(
        "blissoda.utils.trigger.disable_processing_trigger",
        side_effect=disable,
        autospec=True,
    ):
        yield enabled



from blissoda.demo.processors.xrpd_id31 import DemoId31XrpdProcessor  # noqa F401
from blissoda.demo.tests.itest_streamline_id31 import (  # noqa F401
    DemoId31StreamlineScanner,
)
from blissoda.demo.tests.itest_streamline_id31 import (  # noqa F401
    id31_streamline_auto_stop_demo,
)
from blissoda.demo.tests.itest_streamline_id31 import id31_streamline_demo  # noqa F401
from blissoda.demo.tests.itest_xrpd_id31 import id31_xrpd_demo  # noqa F401



import gc
import os

from ewoksjob.client.celery.utils import get_not_finished_futures

from ... import flint
from ...bliss_globals import current_session
from ...bliss_globals import setup_globals
from .. import BLISS_LOGS_DIR
from .. import testing


@testing.integration_fixture(autouse=True)
def cleanup_celery_jobs():
    yield
    futures = get_not_finished_futures()
    if futures:
        testing.print_message(
            f"{len(futures)} workflows are still running: abort them.", "warning"
        )
    for future in futures:
        _ = future.abort()
        try:
            _ = future.result(timeout=5)
        except Exception as ex:
            testing.print_message(f"Aborted workflow: {ex}", "warning")


@testing.integration_fixture
def gc_collect():
    while gc.collect():
        pass
    yield
    while gc.collect():
        pass


@testing.integration_fixture(autouse=True)
def debug_blissoda():
    setup_globals.debugon("blissoda.*")
    yield
    setup_globals.debugoff("blissoda.*")


@testing.integration_fixture
def debug_flint():
    keep = current_session.scan_display.extra_args

    log_file = os.path.join(BLISS_LOGS_DIR, "flint_itest.log")
    extra_args = ["--debug", "--log-file", log_file]

    try:
        flint.FlintClient().close()
    except RuntimeError:
        pass  # already closed

    current_session.scan_display.extra_args = extra_args

    yield

    current_session.scan_display.extra_args = keep

    try:
        flint.FlintClient().close()
    except RuntimeError:
        pass  # already closed



"""Integration tests for the generic EXAFS processor."""

from .. import testing
from ..processors.exafs import DemoExafsProcessor
from .utils import exafs


def exafs_demo(nrepeats: int = None, fast: bool = True):
    test_exafs_scan(nrepeats=nrepeats, fast=fast)


@testing.integration_fixture
def exafs_processor():
    return DemoExafsProcessor()


@testing.integration_test
def test_exafs_scan(exafs_processor, nrepeats: int = None, fast: bool = True):
    if nrepeats is None:
        nrepeats = exafs_processor.max_scans + 1
    exafs_processor.clear()
    for _ in range(nrepeats):
        exafs.test_exafs_scan(exafs_processor, fast=fast, split=True)



from importlib.metadata import version

from packaging.version import Version


def has_minimal_version(distribution: str, minimal_version: str) -> bool:
    """Check if the installed version of a distribution meets or exceeds min_version."""
    return Version(version(distribution)) >= Version(minimal_version)


def assert_has_minimal_version(distribution: str, minimal_version: str) -> None:
    """Raise RuntimeError if the installed version of a distribution is below min_version."""
    if not has_minimal_version(distribution, minimal_version):
        raise RuntimeError(f"Requires {distribution!r} >= {minimal_version}")



import json
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ...bliss_globals import current_session
from ...bliss_globals import setup_globals
from ...id22.xrpd_processor import Id22XrpdProcessor
from ...utils import directories
from ..processors.utils.pyfai_calib import DEFAULT_CALIB
from ..processors.utils.xrpd import lima_image_is_padded


class DemoId22XrpdProcessor(Id22XrpdProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("queue", "celery")
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_cython",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )

        super().__init__(config=config, defaults=defaults)
        if self._HAS_BLISS:
            self._ensure_config_filename()

    def get_integrate_1d_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_integrate_1d_inputs(scan, lima_name)
        inputs.append(
            {
                "task_identifier": "Integrate1D",
                "name": "demo",
                "value": lima_image_is_padded(lima_name),
            }
        )
        return inputs

    def _get_demo_result_dir(self, dataset_filename: str) -> str:
        root_dir = directories.get_processed_dir(dataset_filename)
        return os.path.join(root_dir, "demo", "id22")

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_demo_result_dir(dataset_filename)
        return os.path.join(root_dir, "workflows")

    def _get_config_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_demo_result_dir(dataset_filename)
        return os.path.join(root_dir, "config")

    def _ensure_config_filename(self):
        if self.pyfai_config:
            return
        root_dir = self._get_config_dir(current_session.scan_saving.filename)
        cfgfile = os.path.join(root_dir, "pyfaicalib.json")
        os.makedirs(os.path.dirname(cfgfile), exist_ok=True)
        poni = DEFAULT_CALIB["difflab6"]
        with open(cfgfile, "w") as f:
            json.dump(poni, f)
        self.pyfai_config = cfgfile

    def pct(self, *args, **kw):
        s = setup_globals.ct(*args, **kw)
        self.on_new_scan(s)
        return s

    def _expected_data_shapes_1d(self, lima_name: str, npoints: int):
        return {
            f"{lima_name}:q": (4096,),
            f"{lima_name}:intensity": (4096,),
            f"{lima_name}:intensity_errors": (4096,),
        }



import re

import pytest

from ..xrpd.processor import XrpdProcessor


def test_lima_conflicts(mock_persistent, mock_processing_triggers):
    config = {
        "name": "xrpd_processor1",
        "plugin": "generic",
        "class": "XrpdProcessor",
        "package": "blissoda.xrpd.processor",
        "default_parameters": {"lima_names": ["unknown_detector1"], "_enabled": True},
        "shared": False,
        "singleton": False,
    }
    xrpd_processor1 = XrpdProcessor(config=config)
    assert xrpd_processor1.lima_names == ["unknown_detector1"]
    assert xrpd_processor1.get_all_lima_names() == {"unknown_detector1"}

    config = {
        "name": "xrpd_processor2",
        "plugin": "generic",
        "class": "XrpdProcessor",
        "package": "blissoda.xrpd.processor",
        "default_parameters": {"lima_names": ["unknown_detector2"], "_enabled": True},
        "shared": False,
        "singleton": False,
    }
    xrpd_processor2 = XrpdProcessor(config=config)
    assert xrpd_processor2.lima_names == ["unknown_detector2"]
    assert xrpd_processor1.get_all_lima_names() == {
        "unknown_detector1",
        "unknown_detector2",
    }
    assert xrpd_processor2.get_all_lima_names() == {
        "unknown_detector1",
        "unknown_detector2",
    }

    config = {
        "name": "xrpd_processor3",
        "plugin": "generic",
        "class": "XrpdProcessor",
        "package": "blissoda.xrpd.processor",
        "default_parameters": {"lima_names": ["unknown_detector2"], "_enabled": True},
        "shared": False,
        "singleton": False,
    }

    # Fails to instantiate with conflicting Lima name
    expected_error = re.escape(
        "Lima detectors ['unknown_detector2'] are already processed by another XrpdProcessor"
    )
    with pytest.raises(ValueError, match=expected_error):
        _ = XrpdProcessor(config=config)

    # Succeeds to instantiate because Lima name is no longer used
    del xrpd_processor2
    xrpd_processor3 = XrpdProcessor(config=config)
    assert xrpd_processor3.lima_names == ["unknown_detector2"]
    assert xrpd_processor1.get_all_lima_names() == {
        "unknown_detector1",
        "unknown_detector2",
    }
    assert xrpd_processor3.get_all_lima_names() == {
        "unknown_detector1",
        "unknown_detector2",
    }

    # Fails to set Lima name because of conflict
    expected_error = re.escape(
        "Lima detectors ['unknown_detector1'] are already processed by another XrpdProcessor"
    )
    with pytest.raises(ValueError, match=expected_error):
        xrpd_processor3.lima_names = ["unknown_detector1"]

    # Succeeds to set Lima name because Lima name is no longer used
    del xrpd_processor1
    xrpd_processor3.lima_names = ["unknown_detector1"]
    assert xrpd_processor3.lima_names == ["unknown_detector1"]
    assert xrpd_processor3.get_all_lima_names() == {"unknown_detector1"}

    # Succeeds to instantiate because processor is enabled
    config = {
        "name": "xrpd_processor4",
        "plugin": "generic",
        "class": "XrpdProcessor",
        "package": "blissoda.xrpd.processor",
        "default_parameters": {
            "lima_names": ["unknown_detector1"],
        },
        "shared": False,
        "singleton": False,
    }
    xrpd_processor4 = XrpdProcessor(config=config)
    assert xrpd_processor3.get_all_lima_names() == {"unknown_detector1"}
    assert xrpd_processor4.get_all_lima_names() == {"unknown_detector1"}

    # Failes to enable the processing with conflicting Lima name
    expected_error = re.escape(
        "Lima detectors ['unknown_detector1'] are already processed by another XrpdProcessor"
    )
    with pytest.raises(ValueError, match=expected_error):
        xrpd_processor4.enable()



"""Bliss-side for Flint BM32 XRR plots."""

from ..flint import BasePlot


class Bm32XrrAnalysisPlot(BasePlot):
    WIDGET = "blissoda.bm32.xrr_widgets.Bm32XrrAnalysisWidget"

    def clear_data(self) -> None:
        self.submit("clear_data")

    def set_result_data(self, result: dict, sample_name: str = "") -> None:
        self.submit("set_result_data", result, sample_name)



import os
import shutil
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from ...bliss_globals import current_session
from ...fluo.processor import FluoProcessor
from ...processor import BlissScanType
from ...resources import resource_filename
from ...utils import directories


class DemoFluoProcessor(FluoProcessor):

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = dict()
        defaults.setdefault("instrument_name", "demo")
        # TODO: blissdemo worker cannot handle sub-processes so fit one detector for now
        defaults.setdefault("detectors", ["mca1_det2"])
        super().__init__(config=config, defaults=defaults)
        self._set_demo_pymca_configs()
        self._last_future = None

    def _on_new_scan(self, scan: BlissScanType) -> Tuple[Optional[dict], Optional[Any]]:
        metadata, future = super()._on_new_scan(scan)
        self._last_future = future
        return metadata, future

    def _is_xrfmap(self, scan: BlissScanType) -> bool:
        if not scan.scan_info.get("save"):
            return False
        if scan.scan_info["type"] != "amesh":
            return False
        return True

    def _is_fluoxas(self, scan: BlissScanType) -> bool:
        return False

    def _is_mosaic_xrfmap(self, scan: BlissScanType) -> bool:
        return False

    def _set_demo_pymca_configs(self):
        root_dir = self._get_demo_config_dir(current_session.scan_saving.filename)
        dst_cfgfile = os.path.join(root_dir, "pymca.cfg")
        if os.path.exists(dst_cfgfile):
            return

        src_cfgfile = resource_filename("demo", "pymca.cfg")
        os.makedirs(os.path.dirname(dst_cfgfile), exist_ok=True)
        shutil.copyfile(src_cfgfile, dst_cfgfile)

        # Fit each detector separately
        self.pymca_configs = [dst_cfgfile] * len(self.xrf_names)

    def _get_demo_config_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_demo_result_dir(dataset_filename)
        return os.path.join(root_dir, "config")

    def _get_demo_result_dir(self, dataset_filename: str) -> str:
        root_dir = directories.get_processed_dir(dataset_filename)
        return os.path.join(root_dir, "demo", "fluo")



import time

from ..import_utils import unavailable_module
from .access import WithFlintAccess
from .colors import ColorCycler

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)


class BasePlotter(WithFlintAccess):
    def __init__(self, max_plots) -> None:
        super().__init__()
        self._max_plots = max_plots
        self._tasks = []
        self._color_cycler = ColorCycler(max_colors=max_plots + 1)

    @property
    def number_of_scans(self):
        """Maximum number of scans to be plotted"""
        return self._max_plots

    @number_of_scans.setter
    def number_of_scans(self, value):
        self._max_plots = max(value, 0)
        self._on_color_reset()

    def _on_color_reset(self) -> None:
        self._color_cycler.max_colors = self._max_plots

    def handle_workflow_result(self):
        raise NotImplementedError()

    def _spawn(self, *args, **kw):
        task = gevent.spawn(*args, **kw)
        self._tasks.append(task)
        self.purge_tasks()

    def purge_tasks(self) -> int:
        """Remove references to tasks that have finished."""
        self._tasks = [t for t in self._tasks if t]
        return len(self._tasks)

    def kill_tasks(self) -> int:
        """Kill all tasks."""
        gevent.killall(self._tasks)
        return self.purge_tasks()

    def wait_tasks(self, timeout: int = 10) -> int:
        """Wait for all plot tasks to be finished."""
        deadline = time.time() + timeout
        for task in list(self._tasks):
            timeleft = deadline - time.time()
            if timeleft < 0:
                raise TimeoutError(f"{timeout} seconds")
            task.join(timeout=timeleft)
            if task:
                raise TimeoutError(f"{timeout} seconds")
        return self.purge_tasks()

    def replot(self, **retry_options) -> None:
        raise NotImplementedError()



from typing import Optional


class ColorCycler:
    """Cycles through a predefined hex color palette."""

    def __init__(
        self,
        palette_name: Optional[str] = None,
        max_colors: Optional[int] = None,
    ):
        if palette_name is None:
            palette_name = "tableau_10"
        if palette_name not in _COLOR_PALETTES:
            raise ValueError(f"Unknown palette: {palette_name}")

        colors = _COLOR_PALETTES[palette_name]
        self._colors = colors
        self._next_index = 0
        self._max_colors = max_colors or len(colors)

    @property
    def max_colors(self) -> int:
        return self._max_colors

    @max_colors.setter
    def max_colors(self, max_colors: int) -> None:
        self._max_colors = max(max_colors, 0)
        self.reset()

    def next(self) -> str:
        """Return the next color in the cycle (hex string)."""
        color = self._colors[self._next_index]
        n = min(len(self._colors), self.max_colors)
        self._next_index = (self._next_index + 1) % n
        return color

    def reset(self):
        """Reset the cycle to the beginning."""
        self._next_index = 0


_COLOR_PALETTES = {
    "primary": [
        "#386CB0",  # Blue
        "#CC79A7",  # Pink
        "#66A61E",  # Green
        "#E7298A",  # Magenta
        "#7570B3",  # Purple
        "#E6AB02",  # Mustard
        "#A6761D",  # Brown
        "#66CCEE",  # Light Blue
        "#D55E00",  # Orange
        "#009E73",  # Teal
        "#F0E442",  # Yellow
        "#323232",  # Neutral Gray
    ],
    "dark_mode": [
        "#5B9BD5",  # Sky Blue
        "#ED7D31",  # Orange
        "#A5A5A5",  # Gray
        "#FFC000",  # Gold
        "#70AD47",  # Green
        "#255E91",  # Deep Blue
        "#C0504D",  # Red
        "#8064A2",  # Violet
        "#3B9B95",  # Teal
        "#F27CAE",  # Pink
        "#4472C4",  # Indigo
        "#FFD966",  # Warm Yellow
    ],
    "soft": [
        "#B3CDE3",  # Powder Blue
        "#FBB4AE",  # Soft Coral
        "#CCEBC5",  # Mint Green
        "#DECAE4",  # Lavender
        "#FED9A6",  # Peach
        "#FFFFCC",  # Light Yellow
        "#E5D8BD",  # Sand
        "#FDDAEC",  # Rose
        "#C8E6C9",  # Sage
        "#CCCCFF",  # Periwinkle
        "#FFE6CC",  # Cream
        "#CCFFE5",  # Mint
    ],
    "tol_bright": [
        "#4477AA",  # Blue
        "#66CCEE",  # Cyan
        "#228833",  # Green
        "#CCBB44",  # Yellow
        "#EE6677",  # Red
        "#AA3377",  # Purple
        "#BBBBBB",  # Gray
    ],
    "okabe_ito": [
        "#E69F00",  # Orange
        "#56B4E9",  # Sky Blue
        "#009E73",  # Bluish Green
        "#F0E442",  # Yellow
        "#0072B2",  # Blue
        "#D55E00",  # Vermillion
        "#CC79A7",  # Reddish Purple
    ],
    "tableau_10": [
        "#1F77B4",  # Blue
        "#FF7F0E",  # Orange
        "#2CA02C",  # Green
        "#D62728",  # Red
        "#9467BD",  # Purple
        "#8C564B",  # Brown
        "#E377C2",  # Pink
        "#7F7F7F",  # Gray
        "#BCBD22",  # Olive
        "#17BECF",  # Cyan
    ],
}



"""Integration tests for the BM08 EXAFS processor."""

from .. import testing
from ..processors.exafs_bm08 import DemoBm08ExafsProcessor
from .utils import exafs


def bm08_exafs_demo(nrepeats: int = None, fast: bool = True):
    test_exafs_scan(nrepeats=nrepeats, fast=fast)


@testing.integration_fixture
def bm08_exafs_processor():
    return DemoBm08ExafsProcessor()


@testing.integration_test
def test_exafs_scan(bm08_exafs_processor, nrepeats: int = None, fast: bool = True):
    if nrepeats is None:
        nrepeats = bm08_exafs_processor.max_scans + 1
    bm08_exafs_processor.clear()

    scan_info = {
        "scan_parameters": {"scan_type": "Kscan"},
        "scan_metadata_categories": ["scan_parameters"],
    }

    for _ in range(nrepeats):
        exafs.test_exafs_scan(
            bm08_exafs_processor, fast=fast, scan_info=scan_info, split=False
        )



"""Integration tests for the ID22 XRPD processing."""

from ...bliss_globals import setup_globals
from .. import testing
from ..processors.xrpd_id22 import DemoId22XrpdProcessor
from .utils import xrpd


@testing.integration_fixture
def id22_processor():
    xrpd_processor = DemoId22XrpdProcessor()
    xrpd_processor.enable(setup_globals.difflab6)
    xrpd_processor._plotter.clear_lima_plots("difflab6")

    yield xrpd_processor

    xrpd_processor.disable()


def id22_xrpd_demo(expo=0.2, npoints=10):
    test_id22_scan_with_1d_integration(expo=expo, npoints=npoints)


@testing.integration_test
def test_id22_scan_with_1d_integration(
    debug_flint, id22_processor, nrepeats: int = None, expo=0.2, npoints=10
):
    xrpd.test_xrpd_scan_with_1d_integration(
        [id22_processor],
        [
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        ],
        nrepeats=nrepeats,
        expo=expo,
        npoints=npoints,
        plotting_fails=False,  # TODO: plotting of the sum spectrum does not work
    )



"""Integration tests for the ID09 TXS processing."""

from ...bliss_globals import setup_globals
from .. import testing
from ..processors.id09 import DemoTxsProcessor
from .utils import xrpd


@testing.integration_fixture
def txs_processor(gc_collect):
    xrpd_processor = DemoTxsProcessor()
    xrpd_processor.enable()
    xrpd_processor._plotter.clear_lima_plots(xrpd_processor.detector)
    yield xrpd_processor
    xrpd_processor.disable()


def id09_demo(expo=0.2, npoints=10):
    test_id09_ct_with_1d_integration(expo=expo)
    test_id09_scan_with_1d_integration(expo=expo, npoints=npoints)


@testing.integration_test
def test_id09_ct_with_1d_integration(
    debug_flint, txs_processor, nrepeats: int = None, expo=0.2
):
    xrpd.test_xrpd_ct_with_1d_integration(
        txs_processor,
        [setup_globals.difflab6, setup_globals.diode1, setup_globals.diode2],
        nrepeats=nrepeats,
        expo=expo,
    )


@testing.integration_test
def test_id09_scan_with_1d_integration(
    debug_flint, txs_processor, nrepeats: int = None, expo=0.2, npoints=10
):
    xrpd.test_xrpd_scan_with_1d_integration(
        [txs_processor],
        [
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        ],
        nrepeats=nrepeats,
        expo=expo,
        npoints=npoints,
    )



from typing import Optional

from esrf_pathlib import ESRFPath

from ....bliss_globals import setup_globals
from ....import_utils import unavailable_module
from ... import testing

try:
    from bliss.physics import units
except ImportError as ex:
    units = unavailable_module(ex)


def test_exafs_scan(
    exafs_processor,
    fast: bool = True,
    scan_info: Optional[dict] = None,
    split: bool = True,
):
    # Prepare scan
    if fast:
        e0 = 8950  # eV
        e1 = 9200  # eV
        step_size = 3.0  # eV
    else:
        e0 = 8800  # eV
        e1 = 9600  # eV
        step_size = 0.5  # eV

    npoints = int((e1 - e0) / step_size)
    intervals = npoints - 1

    from_unit = "eV"
    to_unit = exafs_processor.counters["energy_unit"]

    e0 = (e0 * units.ur(from_unit)).to(to_unit).magnitude
    e1 = (e1 * units.ur(from_unit)).to(to_unit).magnitude
    scan = setup_globals.ascan(
        setup_globals.energy,
        e0,
        e1,
        intervals,
        0.01,
        setup_globals.mu,
        scan_info=scan_info,
        run=False,
    )

    # Processor is in charge of the scan lifecycle
    exafs_processor.run(scan)

    # Validate results
    scan_number = scan.scan_info["scan_nb"]
    raw_dataset_file = ESRFPath(scan.scan_info["filename"])

    testing.wait_workflows()
    _assert_flint_plots_exist(
        exafs_processor, f"{raw_dataset_file.stem}: {scan_number}.1"
    )
    if split:
        _assert_files_exist(
            raw_dataset_file.processed_dataset_file, scan_number, npoints
        )


@testing.demo_assert("Check EXAFS Flint plot {scan_label}")
def _assert_flint_plots_exist(exafs_processor, scan_label):
    exafs_plot = exafs_processor._plotter._get_plot()
    scan_labels = exafs_plot.get_scans()

    if len(scan_labels) > exafs_processor.max_scans:
        raise AssertionError(
            f"{len(scan_labels)} scans plotted exceeds the maximum of {exafs_processor.max_scans}"
        )

    if scan_label not in scan_labels:
        raise AssertionError(f"{scan_label!r} not in {scan_labels}")


@testing.demo_assert("Check EXAFS workflow results for scan #{scan_number}")
def _assert_files_exist(filename, scan_number, npoints):
    if not filename.exists():
        raise FileNotFoundError(str(filename))

    testing.assert_hdf5_dataset_exists(
        filename, f"/{scan_number}.1/measurement/mu", (npoints,)
    )
    testing.assert_hdf5_dataset_exists(
        filename, f"/{scan_number}.1/measurement/energy", (npoints,)
    )



from __future__ import annotations

import json
import logging
from pathlib import Path

from ...bliss_globals import current_session  # type: ignore
from ...id31.stop_scan_preset import StopIntegrateSum
from ...utils import directories
from .streamline_id31 import id31_patching
from .utils.pyfai_calib import DEFAULT_CALIB

_logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------
# Minimal Ewoks workflow: PyFaiConfig → IntegrateAndSendMsg
# ---------------------------------------------------------------------------
_DEMO_WORKFLOW = {
    "graph": {"schema_version": "1.1", "id": "stop_scan_pyfai_demo"},
    "nodes": [
        {
            "id": "config",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.pyfaiconfig.PyFaiConfig",
        },
        {
            "id": "integrate",
            "task_type": "class",
            "task_identifier": "blissoda.demo.tasks.stop_scan_xrpd_integrate.IntegrateAndSendMsg",
        },
    ],
    "links": [
        {
            "source": "config",
            "target": "integrate",
            "data_mapping": [
                {
                    "source_output": "detector_config",
                    "target_input": "detector_config",
                },
                {"source_output": "energy", "target_input": "energy"},
                {
                    "source_output": "integration_options",
                    "target_input": "integration_options",
                },
                {"source_output": "detector", "target_input": "detector"},
                {"source_output": "mask", "target_input": "mask"},
                {"source_output": "geometry", "target_input": "geometry"},
            ],
        },
    ],
}


# ---------------------------------------------------------------------------
# DemoStopIntegrateSum
# ---------------------------------------------------------------------------
class DemoStopIntegrateSum(StopIntegrateSum):
    """A demo-friendly ``StopIntegrateSum`` for the Bliss demo session.

    Calls ``id31_patching()`` before ``super().__init__()`` so that:

    * ``setup_globals.atten`` resolves to the mock ``controllers.Attenuator``
      (with ``.bits`` attribute),
    * ``from id31.attenuator import Attenuator, SiO2trans`` resolves to the
      mock implementations.

    The PyFAI config JSON and Ewoks workflow are auto-created in the
    processed directory before the parent is initialised.

    All other behaviour (workflow submission, Redis stream feedback,
    ``ScanPointIterator`` / ``_adjust_filter`` logic) is inherited unchanged.
    """

    def __init__(
        self,
        workflow_threshold: float,
        detector_name: str = "difflab6",
        detector_saturation: float = 1e6,
        frame_target_max: float | None = None,
        frame_target_min: float | None = None,
    ) -> None:
        id31_patching(energy=75.05)  # need it for proper atten tests
        pyfai_config_path = _ensure_demo_pyfai_config()
        workflow_path = _ensure_demo_workflow()

        super().__init__(
            workflow_threshold=workflow_threshold,
            workflow_path=workflow_path,
            detector_name=detector_name,
            detector_saturation=detector_saturation,
            pyfai_config_path=pyfai_config_path,
            attenuator_name="atten",
            frame_target_max=frame_target_max,
            frame_target_min=frame_target_min,
            ewoksjob_queue="celery",
        )


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------
def _ensure_demo_pyfai_config() -> str:
    """Create a demo PyFAI config JSON for ``difflab6`` if it doesn't exist."""
    processed_dir = directories.get_processed_dir(current_session.scan_saving.filename)
    config_path = Path(processed_dir, "config", "pyfaicalib.json")
    if not config_path.is_file():
        _logger.info("Create demo PyFAI config file: %s", config_path)
        config_path.parent.mkdir(parents=True, exist_ok=True)
        config_path.write_text(json.dumps(DEFAULT_CALIB["difflab6"]))
    return str(config_path)


def _ensure_demo_workflow() -> str:
    """Write the demo workflow JSON into the processed directory."""
    processed_dir = directories.get_processed_dir(current_session.scan_saving.filename)
    workflow_path = Path(processed_dir, "workflows", "stop_scan_xrpd_id31.json")
    if not workflow_path.is_file():
        _logger.info("Create demo stop_scan_xrpd_id31 workflow: %s", workflow_path)
        workflow_path.parent.mkdir(parents=True, exist_ok=True)
        workflow_path.write_text(json.dumps(_DEMO_WORKFLOW, indent=2))
    return str(workflow_path)



from ...bliss_globals import setup_globals
from ...demo.processors.bm02 import DemoBm02XrpdProcessor
from ...demo.tests.utils.xrpd import _scan_context
from ...processor import BlissScanType
from .. import testing


@testing.integration_fixture
def bm02_xrpd_processor(gc_collect):
    bm02_processor = DemoBm02XrpdProcessor()
    bm02_processor.enable(setup_globals.difflab6)
    yield bm02_processor
    bm02_processor.disable_empty_cell_subtraction(setup_globals.difflab6)
    bm02_processor.disable()


def _scan(expo=0.2, npoints=10) -> BlissScanType:
    with _scan_context() as scans:
        scan = setup_globals.loopscan(npoints, expo, setup_globals.difflab6)
        scans.append(scan)
        return scan


def bm02_demo_1d(expo=0.2, npoints=10, ascii=False):
    test_bm02_1d(expo=expo, npoints=npoints, ascii=ascii)


@testing.integration_test
def test_bm02_1d(bm02_xrpd_processor, expo=0.2, npoints=10, ascii=False):
    bm02_xrpd_processor.ascii_export_enabled = ascii
    bm02_xrpd_processor.integration_options["difflab6"].pop("nbpt_azim", None)
    _scan(expo, npoints)
    testing.wait_workflows(raise_on_error=True)


def bm02_demo_2d(expo=0.2, npoints=10, ascii=False):
    test_bm02_2d(expo=expo, npoints=npoints, ascii=ascii)


@testing.integration_test
def test_bm02_2d(bm02_xrpd_processor, expo=0.2, npoints=10, ascii=False):
    bm02_xrpd_processor.integration_options["difflab6"]["nbpt_azim"] = 100
    bm02_xrpd_processor.ascii_export_enabled = ascii
    _scan(expo, npoints)
    testing.wait_workflows(raise_on_error=True)


def bm02_demo_1d_with_cell_subtraction(expo=0.2, npoints=10, ascii=False):
    test_bm02_1d_with_cell_subtraction(expo=expo, npoints=npoints, ascii=ascii)


@testing.integration_test
def test_bm02_1d_with_cell_subtraction(
    bm02_xrpd_processor, expo=0.2, npoints=10, ascii=False
):
    bm02_xrpd_processor.ascii_export_enabled = ascii
    bm02_xrpd_processor.integration_options["difflab6"].pop("nbpt_azim", None)

    # Do a first scan to be subtracted
    scan = _scan(expo, npoints)
    testing.wait_workflows(raise_on_error=True)
    scan_metadata = scan.scan_saving

    bm02_xrpd_processor.enable_empty_cell_subtraction(setup_globals.difflab6)
    bm02_xrpd_processor.set_cell_pattern_url(
        sample=scan_metadata.sample_name,
        dataset=scan_metadata.dataset,
        scan_number=scan.scan_info["scan_nb"],
        detector="difflab6",
    )

    _scan(expo, npoints)
    testing.wait_workflows(raise_on_error=True)



from ...bm02.xrpd_processor import Bm02XrpdProcessor
from ...bm08.converter import Bm08Hdf5ToXdiConverter
from ...bm08.exafs_plotter import Bm08ExafsProcessor
from ...bm23.exafs_processor import Bm23ExafsProcessor
from ...exafs.processor import ExafsProcessor
from ...id01.cdi_processor import CdiProcessor
from ...id11.xrpd_processor import Id11XrpdProcessor
from ...id12.converter import Id12Hdf5ToAsciiConverter
from ...id14.converter import Id14Hdf5ToSpecConverter
from ...id22.stscan_processor import StScanProcessor
from ...id22.xrpd_processor import Id22XrpdProcessor
from ...id31.streamline_scanner import Id31StreamlineScanner
from ...id31.xrpd_processor import Id31XrpdProcessor
from ...id32.processor import Id32SpecGenProcessor
from ...streamline.scanner import StreamlineScanner
from ...wrappers.ewoks_macros import EwoksMacroHandler
from ...xrpd.processor import XrpdProcessor


def all_print():
    _print_objects(XrpdProcessor)
    _print_objects(ExafsProcessor)
    _print_objects(StreamlineScanner)
    _print_objects(Bm23ExafsProcessor)
    _print_objects(Bm02XrpdProcessor)
    _print_objects(Id11XrpdProcessor)
    _print_objects(Id22XrpdProcessor)
    _print_objects(Id31XrpdProcessor)
    _print_objects(Id14Hdf5ToSpecConverter)
    _print_objects(Id12Hdf5ToAsciiConverter)
    _print_objects(Id31StreamlineScanner)
    _print_objects(Id32SpecGenProcessor)
    _print_objects(EwoksMacroHandler)
    _print_objects(Bm08Hdf5ToXdiConverter)
    _print_objects(StScanProcessor)
    _print_objects(CdiProcessor)
    _print_objects(Bm08ExafsProcessor)


def _print_objects(processor_cls):
    print()
    print("===================")
    obj = processor_cls()
    print(obj._parameters.name)
    print(obj.__info__())



import os
from glob import glob
from pathlib import Path
from typing import Dict
from typing import List
from typing import Tuple

import h5py

from ....bliss_globals import current_session
from ....bliss_globals import setup_globals
from ....streamline.scanner import StreamlineScanner
from ....utils import directories
from ....version_utils import has_minimal_version
from ... import testing


def test_streamline_without_autocalibration(streamline_scanner: StreamlineScanner):
    # Nothing loaded and 2 holders in the tray:
    streamline_scanner.eject()
    streamline_scanner.sample_changer.reset_tray(2)
    result_dir_pattern, holder_counter = _unused_holder_counter(streamline_scanner)

    # Initialize workflow without calibration
    streamline_scanner.init_workflow(with_autocalibration=False)

    # Measure all holders
    streamline_scanner.run(0.1, setup_globals.difflab6)

    # Verify integrated results
    samples = {
        holder: {sample: [1] for sample in range(16)}
        for holder in range(holder_counter, holder_counter + 2)
    }
    _assert_results(streamline_scanner, result_dir_pattern, samples)


def test_streamline_with_autocalibration(streamline_scanner: StreamlineScanner):
    # Nothing loaded and 1 holder in the tray:
    streamline_scanner.eject()
    streamline_scanner.sample_changer.reset_tray(1)
    result_dir_pattern, holder_counter = _unused_holder_counter(streamline_scanner)

    # Initialize workflow with calibration
    streamline_scanner.init_workflow(with_autocalibration=True)

    # Load, measure and perform calibration
    streamline_scanner.load()
    streamline_scanner.calib(0.1, setup_globals.difflab6, sample_index=0)

    testing.wait_workflows(raise_on_error=True)

    # Verify integrated results
    samples = {holder_counter: {0: [1]}}
    _assert_results(streamline_scanner, result_dir_pattern, samples)

    # Nothing loaded and 2 holders in the tray:
    streamline_scanner.eject()
    streamline_scanner.sample_changer.reset_tray(2)
    result_dir_pattern, holder_counter = _unused_holder_counter(streamline_scanner)

    # Measure all holders
    streamline_scanner.run(0.1, setup_globals.difflab6)

    # Verify integrated results
    samples = {
        holder: {sample: [1] for sample in range(16)}
        for holder in range(holder_counter, holder_counter + 2)
    }
    _assert_results(streamline_scanner, result_dir_pattern, samples)


def _assert_results(
    streamline_scanner,
    result_dir_pattern: str,
    samples: Dict[int, Dict[int, List[int]]],
):
    testing.wait_workflows(raise_on_error=True)

    if streamline_scanner.workflow_has_calib:
        data_paths = streamline_scanner._TEST_DATA_PATHS_WITH_CALIB
        ascii_suffixes = streamline_scanner._TEST_ASCII_SUFFIXES_WITH_CALIB
    else:
        data_paths = streamline_scanner._TEST_DATA_PATHS_WITHOUT_CALIB
        ascii_suffixes = streamline_scanner._TEST_ASCII_SUFFIXES_WITHOUT_CALIB

    for holder_counter, sample in samples.items():
        for sample_index, dataset_indices in sample.items():
            for dataset_index in dataset_indices:
                result_dir = Path(
                    result_dir_pattern.format(
                        holder_counter=holder_counter,
                        sample_index=sample_index,
                        dataset_index=dataset_index,
                    )
                )

                file_path = result_dir / f"{result_dir.name}.h5"
                for data_path in data_paths:
                    testing.assert_hdf5_dataset_exists(file_path, data_path, (4096,))

                for ascii_suffix in ascii_suffixes:
                    ascii_file = file_path.parent / f"{file_path.stem}{ascii_suffix}"
                    assert ascii_file.exists(), f"File '{ascii_file}' does not exist"

                thumbnail = result_dir / "gallery" / "integrate.png"
                assert thumbnail.exists(), f"File '{thumbnail}' does not exist"


def test_streamline_id31_auto_stop(streamline_scanner: StreamlineScanner):
    """Integration test for the "auto-stop" accumulation mode.

    Configures the scanner for auto-stop mode, runs a single sample, and
    verifies that:
    - A ``*_sum.h5`` file is written with a 2D ``sum_frame`` dataset.
    - The normal streamline pipeline produces integrated 1D results.
    """
    if not has_minimal_version("bliss", "2.2"):
        testing.skip_integration_test("Requires bliss>=2.2")

    # Nothing loaded, 1 holder in the tray
    streamline_scanner.eject()
    streamline_scanner.sample_changer.reset_tray(1)
    result_dir_pattern, holder_counter = _unused_holder_counter(streamline_scanner)

    # Enable auto-stop mode
    streamline_scanner.auto_stop_acc_mode = True
    streamline_scanner.auto_stop_threshold = 0.005
    streamline_scanner.optimize_exposure_per = None
    streamline_scanner.queue = "celery"

    # Init the streamline workflow and the auto-stop workflow
    streamline_scanner.init_workflow(with_autocalibration=False)
    streamline_scanner.init_auto_stop_workflow()

    # Measure sample 0 and 1 only
    streamline_scanner.run(0.2, setup_globals.difflab6, sample_indices=[0, 1])

    # Assert streamline pipeline produced integrated results
    samples = {holder_counter: {0: [1]}}
    _assert_results(streamline_scanner, result_dir_pattern, samples)

    # Assert that a *_sum.h5 file was created with a 2D sum_frame dataset.
    # Assert that scan auto stopped early.
    # After run() finishes, scan_saving.filename has advanced to the next dataset,
    # so search one level up (the sample directory) to find the sum file from the
    # completed scan.
    dataset_filename = current_session.scan_saving.filename
    sum_files = list(Path(dataset_filename).parent.parent.rglob("*_sum.h5"))
    assert len(sum_files) > 0, "No *_sum.h5 file found in sample directory"
    with h5py.File(str(sum_files[0]), "r") as f:
        assert "sum_frame" in f, f"'sum_frame' dataset not found in {sum_files[0]}"
        ds = f["sum_frame"]
        assert isinstance(
            ds, h5py.Dataset
        ), f"'sum_frame' is not a dataset in {sum_files[0]}"
        sum_frame = ds[:]
        nframes_ds = f["nframes"]
        assert isinstance(nframes_ds, h5py.Dataset)
        nframes = int(nframes_ds[()])
    assert sum_frame.ndim == 2, f"Expected 2D sum_frame, got shape {sum_frame.shape}"
    assert (
        nframes < 20
    ), f"Expected auto-stop to trigger before 20 frames, got {nframes}"


def _unused_holder_counter(streamline_scanner: StreamlineScanner) -> Tuple[str, int]:
    dataset_filename = current_session.scan_saving.filename

    raw_data = directories.get_raw_dir(dataset_filename)

    sample_changer = streamline_scanner.sample_changer

    # Holders start counting from 1
    holder_counter = sample_changer._holder_counter or 1

    while glob(os.path.join(raw_data, f"holder{holder_counter}_*")):
        holder_counter += 1

    # Holder counter increments every time we load a new holder
    sample_changer._holder_counter = holder_counter - 1

    processed_data = directories.get_processed_dir(dataset_filename)
    subdir = "holder{holder_counter}_sample{sample_index:02d}_lab6_{dataset_index:04d}"
    result_dir_pattern = os.path.join(processed_data, "streamline", subdir)
    return result_dir_pattern, holder_counter



from __future__ import annotations

import functools
import logging

from ..flint.plotter import BasePlotter
from ..utils.serialization import ensure_numpy_pickle_compat
from .xrr_plots import Bm32XrrAnalysisPlot

logger = logging.getLogger(__name__)


def _background_task(method):
    @functools.wraps(method)
    def wrapper(*args, **kw):
        try:
            return method(*args, **kw)
        except Exception as ex:
            logger.error("BM32 XRR plotter task failed (%s)", ex, exc_info=True)
            raise

    return wrapper


class Bm32XrrPlotter(BasePlotter):
    def __init__(self) -> None:
        super().__init__(max_plots=1)
        self._last_plot_name = "XRR analysis"

    def handle_workflow_result(
        self,
        future,
        sample_name: str,
        plot_name: str = "XRR analysis",
    ) -> None:
        if not self._HAS_BLISS:
            return
        self._spawn(self._handle_workflow_result, future, sample_name, plot_name)

    def handle_result(
        self,
        result: dict,
        sample_name: str,
        plot_name: str = "XRR analysis",
    ) -> None:
        self._last_plot_name = plot_name
        result = result if isinstance(result, dict) else {}
        try:
            plot = self._get_plot(plot_name, Bm32XrrAnalysisPlot)
            plot.set_result_data(result, sample_name=sample_name)
        except Exception as ex:
            logger.warning("Could not update BM32 XRR Flint plot: %s", ex)

    @_background_task
    def _handle_workflow_result(
        self,
        future,
        sample_name: str,
        plot_name: str,
    ) -> None:
        ensure_numpy_pickle_compat()
        self.handle_result(
            result=future.result(), sample_name=sample_name, plot_name=plot_name
        )

    def clear(self) -> None:
        try:
            plot = self._get_plot(self._last_plot_name, Bm32XrrAnalysisPlot)
            plot.clear_data()
        except Exception as ex:
            logger.warning("Could not clear BM32 XRR Flint plot: %s", ex)

    def replot(self, **_retry_options) -> None:
        pass



import re
from typing import List
from typing import Set

_BLISS_RE = re.compile(r"^(?P<ctrl>[^:]+):spectrum_det(?P<num>\d+)$")
# Example: "fxb:spectrum_det01"

_BLISS_RE_ALT = re.compile(r"^(?P<ctrl>[^:]+):spectrum:det(?P<num>\d+)$")
# Example: "fx_nano:spectrum:det01"

_HDF5_RE = re.compile(r"^(?P<ctrl>.+)_det(?P<num>\d+)$")
# Example: "fxb_det01"


def counter_to_h5_names(xrf_names: List[str]) -> List[str]:
    """
    Convert counter names to HDF5 NXdetector names.

    Examples:
    - "fxb:spectrum_det01"    -> "fxb_det01"
    - "fxb_det01"             -> "fxb_det01"
    - "fx_nano:spectrum:det00 -> "fx_nano_det0"  # no leading zeros ????
    """
    h5_names = []

    for name in xrf_names:
        if m := _BLISS_RE.match(name):
            prefix = m.group("ctrl")
            number = m.group("num")
            h5_names.append(f"{prefix}_det{number}")

        elif m := _BLISS_RE_ALT.match(name):
            prefix = m.group("ctrl")
            number = int(m.group("num"))
            h5_names.append(f"{prefix}_det{number}")

        elif _HDF5_RE.match(name):
            h5_names.append(name)

        else:
            raise ValueError(f"Unrecognized XRF name format: {name}")

    return h5_names


def counter_to_controller_names(xrf_names: List[str]) -> Set[str]:
    """
    Extract controller names from detector names.

    Examples:
    - "fxb:spectrum_det01"    -> "fxb"
    - "fxb_det01"             -> "fxb"
    - "fx_nano:spectrum:det00 -> "fx_nano"
    """
    controllers: Set[str] = set()

    for name in xrf_names:
        if m := _BLISS_RE.match(name):
            controllers.add(m.group("ctrl"))

        elif m := _BLISS_RE_ALT.match(name):
            controllers.add(m.group("ctrl"))

        elif m := _HDF5_RE.match(name):
            controllers.add(m.group("ctrl"))

        else:
            raise ValueError(f"Unrecognized XRF name format: {name}")

    return controllers



from blissoda.demo.processors.stop_scan_xrpd_id31 import (  # noqa F401
    DemoStopIntegrateSum,
)
from blissoda.demo.tests.itest_stop_scan_xrpd_id31 import (  # noqa F401
    stop_scan_pyfai_demo,
)
from blissoda.demo.tests.itest_stop_scan_xrpd_id31 import (  # noqa: F401
    test_attenuation_decrease,
)
from blissoda.demo.tests.itest_stop_scan_xrpd_id31 import (  # noqa: F401
    test_attenuation_increase,
)



import os.path
from pathlib import Path

import h5py
from silx.io.url import DataUrl

from ..bliss_globals import current_session
from ..processor import BlissScanType
from ..utils.directories import get_filename


def export_filename_prefix(scan: BlissScanType, lima_name: str) -> str:
    scan_filename = Path(get_filename(scan))
    scan_nb = scan.scan_info.get("scan_nb")
    return f"{scan_filename.stem}_s{scan_nb:04d}_{lima_name}"


def get_current_filename() -> str:
    return current_session.scan_saving.filename


def subtracted_nxprocess_name(lima_name: str) -> str:
    return f"{lima_name}_integrate_subtracted"


def assert_url_exists(url: str) -> None:
    data_url = DataUrl(url)
    if not os.path.exists(data_url.file_path()):
        raise FileNotFoundError(f"{data_url.file_path()} does not exist")

    with h5py.File(data_url.file_path(), "r") as h5file:
        if data_url.data_path() not in h5file:
            raise FileNotFoundError(
                f"{data_url.data_path()} does not exist in {data_url.file_path()}"
            )



import pathlib
from typing import Any
from typing import Tuple

import h5py
import numpy
from ewoksutils import uri_utils
from silx.io.dictdump import h5todict
from silx.io.h5py_utils import retry
from silx.utils.retry import RetryError

from ._assert import demo_assert


@demo_assert("HDF5 group {data_path} in {file_path}")
@retry(retry_timeout=10, retry_period=0.2)
def assert_hdf5_group_exists(file_path: pathlib.Path, data_path: str) -> None:
    if not file_path.exists():
        raise RetryError(f"{str(file_path)!r} does not exist")
    file_path = str(file_path)
    with h5py.File(file_path, mode="r") as root:
        if data_path not in root:
            raise RetryError(f"{data_path!r} not in {file_path!r}")


@demo_assert("HDF5 dataset {data_path} in {file_path} with shape {expected_shape}")
@retry(retry_timeout=10, retry_period=0.2)
def assert_hdf5_dataset_exists(
    file_path: pathlib.Path,
    data_path: str,
    expected_shape: Tuple[int, ...],
) -> None:
    if not file_path.exists():
        raise RetryError(f"{str(file_path)!r} does not exist")
    file_path = str(file_path)
    with h5py.File(file_path, mode="r") as root:
        if data_path not in root:
            raise RetryError(f"{data_path!r} not in {file_path!r}")

        dset = root[data_path]
        if dset.shape != expected_shape:
            url = f"{file_path!r}::{data_path}"
            raise RetryError(
                f"Shape of {url!r} is {dset.shape} instead of {expected_shape}"
            )


@demo_assert("HDF5 uri {uri} exists")
@retry(retry_timeout=10, retry_period=0.2)
def assert_equal_hdf5_content(uri: str, expected: Any):
    file_path = uri_utils.path_from_uri(uri)
    query = uri_utils.parse_query(uri)
    data_path = query["path"]

    if not file_path.exists():
        raise RetryError(f"{str(file_path)!r} does not exist")
    file_path = str(file_path)
    with h5py.File(file_path, mode="r") as root:
        if data_path not in root:
            raise RetryError(f"{data_path!r} not in {file_path!r}")
        actual = h5todict(root, data_path)
        __assert_equal(actual, expected)


def __assert_equal(actual: Any, expected: Any, _path: str = "") -> None:
    if isinstance(expected, dict):
        err_msg = f"Expected dict at {_path or '/'} got {type(actual).__name__}"
        assert isinstance(actual, dict), err_msg

        actual_keys = set(actual)
        expected_keys = set(expected)
        if actual_keys != expected_keys:
            missing = expected_keys - actual_keys
            extra = actual_keys - expected_keys
            raise AssertionError(
                f"Key mismatch at {_path or '/'}: " f"missing={missing}, extra={extra}"
            )

        for key in expected:
            next_path = f"{_path}/{key}" if _path else f"/{key}"
            __assert_equal(actual[key], expected[key], _path=next_path)

    elif isinstance(actual, numpy.ndarray):
        err_msg = f"Array differ at {_path!r}"
        numpy.testing.assert_array_equal(actual, expected, err_msg=err_msg)

    else:
        err_msg = (
            f"Values differ at {_path!r}: actual={actual!r}, expected={expected!r}"
        )
        assert actual == expected, err_msg



"""Integration tests for the generic XRPD processor."""

from ...bliss_globals import setup_globals
from .. import testing
from ..processors.xrpd import DemoXrpdProcessor
from .utils import xrpd


def xrpd_demo_1d(nrepeats: int = None, expo=0.2, npoints=10):
    test_xrpd_ct_with_1d_integration(nrepeats=nrepeats, expo=expo)
    test_xrpd_scan_with_1d_integration(nrepeats=nrepeats, expo=expo, npoints=npoints)
    test_xrpd_scan_with_two_1d_integrations(
        nrepeats=nrepeats, expo=expo, npoints=npoints
    )


def xrpd_demo_2d(nrepeats: int = None, expo=0.2, npoints=10):
    test_xrpd_ct_with_2d_integration(nrepeats=nrepeats, expo=expo)
    test_xrpd_scan_with_2d_integration(nrepeats=nrepeats, expo=expo, npoints=npoints)


@testing.integration_fixture
def difflab6_processor_1d(gc_collect):
    xrpd_processor = DemoXrpdProcessor()
    xrpd_processor.enable(setup_globals.difflab6)
    xrpd_processor._plotter.clear_lima_plots("difflab6")
    yield xrpd_processor
    xrpd_processor.disable()


@testing.integration_fixture
def diffcam_processor_1d(gc_collect):
    # Ensure this processor can exist in parallel to difflab6_processor_1d
    config = {
        "name": "diffcam_processor",
        "plugin": "generic",
        "class": "DemoXrpdProcessor",
        "package": "blissoda.demo.processors.xrpd.processor",
        "shared": False,
        "singleton": False,
    }
    xrpd_processor = DemoXrpdProcessor(config=config)
    xrpd_processor.enable(setup_globals.diffcam)
    # TODO: produce different data or fix PONI
    # xrpd_processor.integration_options["mask"] = numpy.zeros((540, 516), dtype=int).tolist()
    xrpd_processor._plotter.clear_lima_plots("diffcam")
    yield xrpd_processor
    xrpd_processor.disable()


@testing.integration_fixture
def difflab6_processor_2d(gc_collect):
    xrpd_processor = DemoXrpdProcessor()
    xrpd_processor.enable(setup_globals.difflab6)
    xrpd_processor.integration_options["nbpt_azim"] = 360
    xrpd_processor._plotter.clear_lima_plots("difflab6")
    yield xrpd_processor
    xrpd_processor.integration_options.pop("nbpt_azim", None)
    xrpd_processor.disable()


@testing.integration_test
def test_xrpd_ct_with_1d_integration(
    debug_flint, difflab6_processor_1d, nrepeats: int = None, expo=0.2
):
    xrpd.test_xrpd_ct_with_1d_integration(
        difflab6_processor_1d,
        [setup_globals.difflab6, setup_globals.diode1, setup_globals.diode2],
        nrepeats=nrepeats,
        expo=expo,
    )


@testing.integration_test
def test_xrpd_ct_with_2d_integration(
    debug_flint, difflab6_processor_2d, nrepeats: int = None, expo=0.2
):
    xrpd.test_xrpd_ct_with_2d_integration(
        difflab6_processor_2d,
        [setup_globals.difflab6, setup_globals.diode1, setup_globals.diode2],
        nrepeats=nrepeats,
        expo=expo,
    )


@testing.integration_test
def test_xrpd_scan_with_1d_integration(
    debug_flint, difflab6_processor_1d, nrepeats: int = None, expo=0.2, npoints=10
):
    xrpd.test_xrpd_scan_with_1d_integration(
        [difflab6_processor_1d],
        [
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        ],
        nrepeats=nrepeats,
        expo=expo,
        npoints=npoints,
    )


@testing.integration_test
def test_xrpd_scan_with_2d_integration(
    debug_flint, difflab6_processor_2d, nrepeats: int = None, expo=0.2, npoints=10
):
    xrpd.test_xrpd_scan_with_2d_integration(
        [difflab6_processor_2d],
        [
            setup_globals.difflab6,
            setup_globals.diode1,
            setup_globals.diode2,
        ],
        nrepeats=nrepeats,
        expo=expo,
        npoints=npoints,
    )


@testing.integration_test
def test_xrpd_scan_with_two_1d_integrations(
    debug_flint,
    difflab6_processor_1d,
    diffcam_processor_1d,
    nrepeats: int = None,
    expo=0.2,
    npoints=10,
):
    xrpd.test_xrpd_scan_with_1d_integration(
        [difflab6_processor_1d, diffcam_processor_1d],
        [
            setup_globals.difflab6,
            setup_globals.diffcam,
            setup_globals.diode1,
            setup_globals.diode2,
        ],
        nrepeats=nrepeats,
        expo=expo,
        npoints=npoints,
    )



from typing import List
from typing import Optional
from typing import Tuple

from marshmallow import fields

from ..daiquiri.mixin import DaiquiriProcessorMixin
from ..daiquiri.mixin import ExposedParameter
from ..daiquiri.validators import exists_list
from ..persistent.parameters import ParameterInfo
from ..resources import resource_filename
from ..utils import directories
from .processor import BlissScanType
from .processor import FluoProcessor


class DaiquiriFluoProcessor(
    FluoProcessor,
    DaiquiriProcessorMixin,
    parameters=[ParameterInfo("notification_queue", category="daiquiri")],
):
    DEFAULT_WORKFLOW: Optional[str] = resource_filename(
        "fluo", "daiquiri_xrfmap_single_detector.json"
    )

    EXPOSED_PARAMETERS = [
        ExposedParameter(
            parameter="workflow",
            field_type=fields.Str,
            title="Last workflow",
            read_only=True,
        ),
        ExposedParameter(
            parameter="pymca_configs",
            field_type=fields.List,
            title="PyMCA Config File(s)",
            field_options={"validate": exists_list},
            field_args=[fields.Str()],
        ),
        ExposedParameter(
            parameter="quantification",
            field_type=fields.Bool,
            title="Quantification",
        ),
    ]

    def scan_requires_processing(self, scan) -> bool:
        """Only execute this workflow if there is a `daiquiri_datacollectionid` in the `scan_info`"""
        if not super().scan_requires_processing(scan):
            return False
        if not scan.scan_info.get("daiquiri_datacollectionid"):
            return False
        return True

    def get_daiquiri_inputs(self, scan: BlissScanType, workflow: str):
        return [
            {
                "task_identifier": "StartJob",
                "name": "dataCollectionId",
                "value": scan.scan_info.get("daiquiri_datacollectionid"),
            },
            {
                "task_identifier": "StartJob",
                "name": "output_filename",
                "value": directories.master_output_filename(scan),
            },
            {
                "task_identifier": "StartJob",
                "name": "workflow",
                "value": workflow,
            },
            {
                "task_identifier": "StartJob",
                "name": "config",
                "value": {
                    "pymca_configs": ",".join(self.pymca_configs),
                },
            },
            {
                "task_identifier": "NotifyBeamline",
                "name": "beamline",
                "value": self.notification_queue,
            },
        ]

    def get_submit_arguments(
        self, scan: BlissScanType, inputs: List[dict]
    ) -> Tuple[str, dict]:
        """Merge in daiquiri StartJob options and use the ppf engine for error handling."""
        workflow, kwargs = super().get_submit_arguments(scan, inputs)
        kwargs["inputs"] += self.get_daiquiri_inputs(scan, workflow)
        kwargs["engine"] = "ppf"
        return workflow, kwargs

    def _get_workflow(self, scan: BlissScanType, workflow_stem: str) -> Optional[str]:
        """Use the daiquiri variant with Mimosa tasks."""
        return resource_filename("fluo", f"daiquiri_{workflow_stem}.json")



from blissoda.bliss_globals import setup_globals

from .itest_bm02 import bm02_demo_1d
from .itest_bm02 import bm02_demo_1d_with_cell_subtraction
from .itest_bm02 import bm02_demo_2d
from .itest_bm08 import bm08_demo
from .itest_ewoks_macros import ewoks_macro_demo
from .itest_exafs import exafs_demo
from .itest_exafs_bm08 import bm08_exafs_demo
from .itest_fluo import fluo_demo
from .itest_id01 import id01_cdi_demo
from .itest_id09 import id09_demo
from .itest_id11 import id11_demo
from .itest_id12 import id12_demo
from .itest_id14 import id14_demo
from .itest_id22 import id22_stscan_demo
from .itest_id32 import id32_specgen_demo
from .itest_stop_scan import stop_scan_demo
from .itest_stop_scan_xrpd_id31 import stop_scan_pyfai_demo
from .itest_streamline import streamline_demo
from .itest_streamline_id31 import id31_streamline_demo
from .itest_template import template_demo
from .itest_xrpd import xrpd_demo_1d
from .itest_xrpd import xrpd_demo_2d
from .itest_xrpd_id22 import id22_xrpd_demo
from .itest_xrpd_id31 import id31_xrpd_demo


def all_demo():
    print()
    print("===================")
    template_demo()

    print()
    print("===================")
    setup_globals.newcollection("xrpd_1d_collection")
    xrpd_demo_1d()

    print()
    print("===================")
    setup_globals.newcollection("xrpd_2d_collection")
    xrpd_demo_2d()

    print()
    print("===================")
    setup_globals.newcollection("streamline_collection")
    streamline_demo()

    print()
    print("===================")
    setup_globals.newcollection("streamline_id31_collection")
    id31_streamline_demo()

    print()
    print("===================")
    setup_globals.newcollection("bm02_collection")
    bm02_demo_1d()
    bm02_demo_1d_with_cell_subtraction()
    bm02_demo_2d()

    print()
    print("===================")
    setup_globals.newcollection("id22_stscan_collection")
    id22_stscan_demo()

    print()
    print("===================")
    setup_globals.newcollection("id22_xrpd_collection")
    id22_xrpd_demo()

    print()
    print("===================")
    setup_globals.newcollection("id11_collection")
    id11_demo()

    print()
    print("===================")
    setup_globals.newcollection("macro_collection")
    ewoks_macro_demo()

    print()
    print("===================")
    setup_globals.newcollection("id14_collection")
    id14_demo()

    print()
    print("===================")
    setup_globals.newcollection("id12_collection")
    id12_demo()

    print()
    print("===================")
    setup_globals.newcollection("bm08_collection")
    bm08_demo()

    print()
    print("===================")
    setup_globals.newcollection("fluo_collection")
    fluo_demo()

    print()
    print("===================")
    setup_globals.newcollection("id09_collection")
    id09_demo()

    print()
    print("===================")
    setup_globals.newcollection("id31_collection")
    id31_xrpd_demo()

    print()
    print("===================")
    setup_globals.newcollection("id01_collection")
    id01_cdi_demo()

    print()
    print("===================")
    setup_globals.newcollection("id32_collection")
    id32_specgen_demo()

    # TODO: ewoksxrpd needs to be imported before xraylarch?
    #       Cannot load backend 'Qt5Agg' which requires the 'qt' interactive framework, as 'headless' is currently running
    print()
    print("===================")
    setup_globals.newcollection("exafs_collection")
    exafs_demo()

    print()
    print("===================")
    setup_globals.newcollection("bm08_exafs_collection")
    bm08_exafs_demo()

    print()
    print("===================")
    setup_globals.newcollection("stop_scan_pyfai_collection")
    stop_scan_pyfai_demo()

    print()
    print("===================")
    setup_globals.newcollection("stop_scan_collection")
    stop_scan_demo()

    print()
    print("===================")
    setup_globals.newcollection("done_collection")

    print()
    print("SUCCESS: all demos can be executed")



import logging
from typing import Generator

from blissdata.beacon.data import BeaconData
from blissdata.redis_engine.store import DataStore
from esrf_pathlib import ESRFPath
from ewoksjob.client import submit
from ewoksutils.task_utils import task_inputs

from ...import_utils import unavailable_class
from ...import_utils import unavailable_module
from ...version_utils import assert_has_minimal_version

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)

try:
    from bliss.scanning.chain import AcquisitionChain
except ImportError as ex:
    AcquisitionChain = unavailable_class(ex)


try:
    from bliss.scanning.chain import ChainPreset
except ImportError as ex:
    ChainPreset = unavailable_class(ex)

try:
    from bliss.scanning.chain import ChainIterationPreset
except ImportError as ex:
    ChainIterationPreset = unavailable_class(ex)

logger = logging.getLogger(__name__)


class DemoPreset(ChainPreset):
    """Start an Ewoks workflow at the start of the scan. This workflow analyses
    scan data and provides feedback in the form of events in a Redis stream.

    Each scan point the events are checked with the following result:

    - stop the scan
    - do nothing
    """

    def __init__(self, lima_limits: dict, scalar_counter_limits: dict) -> None:
        assert_has_minimal_version("bliss", "2.2")

        self._lima_limits = lima_limits
        self._scalar_counter_limits = scalar_counter_limits
        self._future = None

        self._filename = None
        self._scan_number = None
        self._scan_key = None

        self._redis_url = BeaconData().get_redis_data_db()
        self._data_store = DataStore(self._redis_url)
        self._stream_name = "blissoda:feedback:stop_task"
        self._last_id = "0-0"

        self._events_buffer = []
        self._greenlet = None

    def start(self, chain: AcquisitionChain) -> None:
        self._cache_scan_info(chain)
        self._start_workflow()

    def stop(self, chain: AcquisitionChain) -> None:
        self._stop_workflow()

    def _cache_scan_info(self, chain: AcquisitionChain) -> None:
        self._filename = chain.scan.scan_info["filename"]
        self._scan_number = chain.scan.scan_info["scan_nb"]
        self._scan_key = chain.scan._scan_data.key

    def _start_workflow(self) -> None:
        """Start the Ewoks workflow and the background greenlet."""
        if self._filename is None:
            raise RuntimeError("Scan file name is not known")

        workflow = {
            "graph": {"schema_version": "1.1", "id": "stop_scan"},
            "nodes": [
                {
                    "task_type": "class",
                    "task_identifier": "blissoda.demo.tasks.stop_scan.StopScan",
                }
            ],
        }

        inputs = task_inputs(
            task_identifier="StopScan",
            inputs=dict(
                filename=self._filename,
                scan_number=self._scan_number,
                scan_key=self._scan_key,
                lima_limits=self._lima_limits,
                scalar_counter_limits=self._scalar_counter_limits,
                redis_url=self._redis_url,
                response_stream_name=self._stream_name,
                retry_timeout=60,
            ),
        )

        # clear previous stream
        self._data_store._redis.delete(self._stream_name)
        self._last_id = "0-0"

        # useful for manual workflow execution during development phase
        convert_destination = str(
            ESRFPath(self._filename).nobackup_path / "last_stop_scan.json"
        )

        # submit workflow
        self._future = submit(
            args=(workflow,),
            kwargs=dict(inputs=inputs, convert_destination=convert_destination),
        )
        logger.debug("Stop scan workflow started: %r", self._future.uuid)

        # start background greenlet
        self._greenlet = gevent.spawn(self._background_event_collector)

    def _stop_workflow(self) -> None:
        """Kill the background greenlet and ensure workflow is finished."""
        if self._greenlet is not None:
            self._greenlet.kill(block=True)
            self._greenlet = None

        # ensure workflow finished
        if self._future is None:
            return
        if self._future.done():
            try:
                self._future.result()
            except Exception as ex:
                logger.error("Stop scan workflow failed: %s", ex)
        else:
            self._future.cancel()

    def get_iterator(
        self, chain: AcquisitionChain
    ) -> Generator[ChainIterationPreset, None, None]:
        """Consume feedback events each iteration."""
        while True:
            yield ChainIterationPreset()

            # consume and clear buffered events
            events_to_process = list(self._events_buffer)
            self._events_buffer.clear()

            for event in events_to_process:
                if event.get(b"type") == b"stop":
                    chain.scan.stop(wait=False)
                    break

    def _background_event_collector(self) -> None:
        """Continuously fetch and buffer feedback events."""
        while True:
            try:
                data = self._data_store._redis.xread(
                    {self._stream_name: self._last_id},
                    count=None,  # read all available
                    block=1000,  # wait for 1 sec until at least one new event
                )

                if not data:
                    continue

                for _, events in data:
                    for event_id, event in events:
                        self._last_id = event_id
                        self._events_buffer.append(event)

            except Exception as ex:
                logger.exception("Error reading Redis stream: %s", ex)



import logging
import os
import re
import shutil
from enum import Enum
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

from ewoksjob.client import get_future
from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils import counters
from ..utils import directories
from .parameters.fluoxas import fluoxas_workflow_inputs
from .parameters.mosaic_xrfmap import mosaic_xrfmap_workflow_inputs
from .parameters.utils import detectors
from .parameters.xrfmap import xrfmap_workflow_inputs

logger = logging.getLogger(__name__)


class ProcessingType(Enum):
    xrfmap = auto()
    fluoxas = auto()
    mosaic_xrfmap = auto()


def _as_list(value):
    if isinstance(value, (str, int)):
        return [value]
    return list(value)


class FluoProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo(
            "workflow",
            category="workflows",
            validator=str,
            doc="Workflow to submit for every scan with MCA detectors.",
        ),
        ParameterInfo(
            "queue",
            category="workflows",
            validator=str,
            doc="Ewoks queue to submit the jobs to.",
        ),
        ParameterInfo(
            "instrument_name",
            category="workflows",
            validator=str,
            doc="Ewoks queue to submit the jobs to.",
        ),
        ParameterInfo(
            "detectors",
            category="PyMCA",
            validator=_as_list,
            deprecated_names=["xrf_names"],
            doc="MCA detectors defined by Bliss channel name, HDF5 name or number",
        ),
        ParameterInfo(
            "pymca_configs",
            category="PyMCA",
            validator=_as_list,
            doc="One configuration file per detector on a single one to fit the sum",
        ),
        ParameterInfo(
            "energy_name",
            category="PyMCA",
            validator=str,
            doc="Name of the primary beam energy counter",
        ),
        ParameterInfo(
            "quantification",
            category="PyMCA",
            validator=bool,
            doc="PyMca quantification",
        ),
        ParameterInfo(
            "norm_counter_name",
            category="Normalization",
            validator=str,
            doc="Name of the counter for normalization",
        ),
        ParameterInfo(
            "virtual_axes",
            category="regrid",
            validator=dict,
            doc="Virtual positioners which are typically the sum of scalar and moving motor positions.",
        ),
        ParameterInfo(
            "ignore_axes",
            category="regrid",
            validator=bool,
            doc="Positioners to ignore when detecting the map motors",
        ),
        ParameterInfo(
            "output_suffix",
            category="results",
            validator=str,
            doc="Output file name suffix",
        ),
        ParameterInfo(
            "data_portal_upload",
            category="data portal",
            validator=bool,
            doc="Upload results to the ESRF Data Portal",
        ),
    ],
):
    """A class that holds parameters related to online workflow triggering for PyMCA."""

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("workflow", "")
        defaults.setdefault("detectors", [])
        defaults.setdefault("output_suffix", "")
        defaults.setdefault("quantification", True)
        defaults.setdefault("virtual_axes", {})
        defaults.setdefault("ignore_axes", [])
        defaults.setdefault("data_portal_upload", False)
        defaults.setdefault("trigger_at", "END")

        super().__init__(config=config, defaults=defaults)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> Optional[dict]:
        metadata, _ = self._on_new_scan(scan)
        return metadata

    def _on_new_scan(self, scan: BlissScanType) -> Tuple[Optional[dict], Optional[Any]]:
        """Executed at the end of every scan"""
        future = None
        metadata = None

        if not self.scan_requires_processing(scan):
            return metadata, future

        # Submit arguments
        processing_type = self._get_processing_type(scan)
        if processing_type is None:
            return metadata, future
        workflow, kwargs = self.get_submit_arguments(scan, processing_type)

        with_saving = scan.scan_info.get("save")
        if with_saving:
            kwargs["convert_destination"] = self._get_workflow_destination(scan)

        upload_parameters = None
        if self.data_portal_upload and with_saving:
            # It is allowed to upload the same processed directory more than once.
            upload_parameters = self._get_workflow_upload_parameters(scan)
        if upload_parameters:
            kwargs["upload_parameters"] = upload_parameters

        # Trigger workflow from the current process.
        logger.info("Starting workflow %r", workflow)
        future = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
        future = get_future(future.uuid)

        return metadata, future

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        """Check whether raw data is saved and whether it contains MCA spectra."""
        if not scan.scan_info.get("save"):
            return False
        if not self._has_mca_controllers(scan):
            return False
        return True

    def get_submit_arguments(
        self, scan: BlissScanType, processing_type: ProcessingType
    ) -> Tuple[str, dict]:
        if processing_type == ProcessingType.xrfmap:
            kwargs = self.get_xrfmap_arguments(scan)
            workflow_stem, inputs = xrfmap_workflow_inputs(**kwargs)
        elif processing_type == ProcessingType.fluoxas:
            kwargs = self.get_fluoxas_arguments(scan)
            workflow_stem, inputs = fluoxas_workflow_inputs(**kwargs)
        elif processing_type == ProcessingType.mosaic_xrfmap:
            kwargs = self.get_mosaic_xrfmap_arguments(scan)
            workflow_stem, inputs = mosaic_xrfmap_workflow_inputs(**kwargs)
        else:
            raise ValueError("Unknown FLUO processing type")

        workflow = self.get_workflow(scan, workflow_stem)
        kwargs = {"inputs": inputs, "outputs": [{"all": False}]}
        return workflow, kwargs

    def _set_workflow(self, scan: BlissScanType, filename) -> None:
        """Set the workflow filename for the scan"""
        self.workflow = filename

    def get_filename(self, scan: BlissScanType) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def _get_axis_units_from_scan(
        self, virtual_axes: Dict[str, str], scan: BlissScanType
    ) -> Optional[Dict[str, str]]:
        units = scan.scan_info["positioners"]["positioners_units"]
        units_dict = {}
        for virtual_axis in virtual_axes:
            for motor_name in _get_motors_from_expression(virtual_axis):
                motor_unit = units[motor_name]
                if motor_unit is None:
                    logger.warning(
                        f"Unit for motor `{motor_name}` is not set in beacon!"
                    )
                else:
                    units_dict[motor_name] = motor_unit
        return units_dict

    def get_xrfmap_arguments(self, scan: BlissScanType) -> dict:
        return dict(
            filename=self.get_filename(scan),
            output_root_uri=self.master_output_url(scan),
            scan_number=scan.scan_info.get("scan_nb"),
            config_filenames=self.pymca_configs,
            instrument_name=self.instrument_name,
            detectors=self.h5_detector_names,
            counter_name=self.norm_counter_name,
            energy_name=self.energy_name,
            quantification=self.quantification,
            virtual_axes=self.virtual_axes,
            ignore_axes=self.ignore_axes,
            axis_units=self._get_axis_units_from_scan(self.virtual_axes.values(), scan),
        )

    def get_mosaic_xrfmap_arguments(self, scan: BlissScanType) -> dict:
        concat_filename = directories.master_output_filename(scan).replace(
            ".h5", "_concat.h5"
        )
        scan_ranges = _get_scan_ranges_from_sequence(scan)

        return dict(
            filenames=[self.get_filename(scan)],
            output_root_uri=self.master_output_url(scan),
            scan_ranges=[scan_ranges],
            concat_bliss_scan_uri=f"{concat_filename}::/{scan_ranges[0]}.1",
            config_filenames=self.pymca_configs,
            instrument_name=self.instrument_name,
            detectors=self.h5_detector_names,
            counter_name=self.norm_counter_name,
            energy_name=self.energy_name,
            quantification=self.quantification,
            virtual_axes=self.virtual_axes,
            ignore_axes=self.ignore_axes,
            axis_units=self._get_axis_units_from_scan(self.virtual_axes.values(), scan),
        )

    def get_fluoxas_arguments(self, scan: BlissScanType) -> dict:
        return dict(
            filenames=[self.get_filename(scan)],
            output_root_uri=self.master_output_url(scan),
            scan_ranges=[_get_scan_ranges_from_sequence(scan)],
            config_filenames=self.pymca_configs,
            instrument_name=self.instrument_name,
            detectors=self.h5_detector_names,
            counter_name=self.norm_counter_name,
            energy_name=self.energy_name,
            quantification=self.quantification,
            virtual_axes=self.virtual_axes
            or None,  # use None when empty (fallback to defaults)
            ignore_axes=self.ignore_axes
            or None,  # use None when empty (fallback to defaults)
            axis_units=self._get_axis_units_from_scan(self.virtual_axes.values(), scan),
        )

    def _get_processing_type(self, scan: BlissScanType) -> Optional[ProcessingType]:
        # Note: we already know the scan contains MCA spectra so no need to check for that
        if self._is_xrfmap(scan):
            return ProcessingType.xrfmap
        if self._is_fluoxas(scan):
            return ProcessingType.fluoxas
        if self._is_mosaic_xrfmap(scan):
            return ProcessingType.mosaic_xrfmap
        return None

    def _is_xrfmap(self, scan: BlissScanType) -> bool:
        return bool(scan.scan_info.get("is_xrfmap"))

    def _is_fluoxas(self, scan: BlissScanType) -> bool:
        return bool(scan.scan_info.get("is_fluoxas"))

    def _is_mosaic_xrfmap(self, scan: BlissScanType) -> bool:
        return bool(scan.scan_info.get("is_xrfmap_patch"))

    def get_workflow(self, scan: BlissScanType, workflow_stem: str) -> Optional[str]:
        """Get the workflow to execute for the scan and ensure it is located
        in the proposal directory for user reference and worker accessibility.
        """
        src_file = self._get_workflow(scan, workflow_stem)
        if src_file is None:
            return

        dataset_filename = self.get_filename(scan)
        workflow_directory = self._get_workflows_dir(dataset_filename)
        dst_file = os.path.join(workflow_directory, os.path.basename(src_file))
        if src_file != dst_file:
            self._set_workflow(scan, dst_file)

        if not os.path.exists(dst_file):
            os.makedirs(workflow_directory, exist_ok=True)
            shutil.copyfile(src_file, dst_file)

        return dst_file

    def _get_workflow(self, scan: BlissScanType, workflow_stem: str) -> Optional[str]:
        return resource_filename("fluo", workflow_stem + ".json")

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        """Directory where to save the un-parameterized workflow."""
        return directories.get_workflows_dir(dataset_filename)

    def _get_workflow_destination(self, scan: BlissScanType) -> str:
        """Directory where to save the workflow parameterized for each scan."""
        convert_destination = directories.workflow_destination(scan)
        dirname, basename = os.path.split(convert_destination)
        base, ext = os.path.splitext(basename)
        return os.path.join(
            dirname, "workflows", f"{base}{self.output_suffix or ''}{ext}"
        )

    @property
    def detector_names(self) -> List[str]:
        """Bliss channel names or HDF5 names. Detector numbers are resolved as strings."""
        return detectors.get_detector_names(self.detectors)

    @property
    def h5_detector_names(self) -> List[str]:
        """HDF5 names only."""
        return counters.counter_to_h5_names(self.detector_names)

    @property
    def xrf_names(self) -> List[str]:
        """Bliss channel names or HDF5 names. Detector numbers are resolved as strings."""
        return self.detector_names

    @xrf_names.setter
    def xrf_names(self, value) -> None:
        """Bliss channel names, HDF5 names or detector numbers."""
        self.detectors = value

    def add_xrf_names(self, detector, *spectra: Sequence[int]) -> None:
        """Add an xrf detector and its spectra to the processor"""
        detector_names = set(self.detector_names)
        for counter in detector.counters:
            if counter.name.startswith("spectrum_det"):
                spectrum_id = int(counter.name.replace("spectrum_det", ""))
                if spectra is None or spectrum_id in spectra:
                    detector_names.add(counter.fullname)
        self.detectors = sorted(detector_names)

    def clear_xrf_names(self):
        self.detectors = []

    def get_xrf_names(self, scan: BlissScanType) -> List[str]:
        # If this is a sequence check the first real scan
        if scan.scan_info.get("is_scan_sequence"):
            channels = scan.streams["SUBSCANS"][0].info.get("channels", dict())
        else:
            channels = scan.scan_info.get("channels", dict())

        return sorted(
            (
                detector_name
                for detector_name in self.detector_names
                if detector_name in channels
            )
        )

    def master_output_url(self, scan: BlissScanType) -> str:
        """URL which can be used to inspect the results after the processing."""
        scan_nb = scan.scan_info.get("scan_nb")
        filename = directories.master_output_filename(scan)
        base, ext = os.path.splitext(filename)
        return f"{base}{self.output_suffix or ''}{ext}::/{scan_nb}.1"

    def _get_workflow_upload_parameters(self, scan: BlissScanType) -> dict:
        raw_directory = os.path.dirname(self.get_filename(scan))
        processed_directory = directories.scan_processed_directory(scan)
        scan_saving = current_session.scan_saving
        metadata = {"Sample_name": scan_saving.dataset["Sample_name"]}

        return {
            "beamline": scan_saving.beamline,
            "proposal": scan_saving.proposal_name,
            "dataset": "xrf_fit",
            "path": processed_directory,
            "raw": [raw_directory],
            "metadata": metadata,
        }

    def _has_mca_controllers(self, scan: BlissScanType) -> bool:
        controller_names = counters.counter_to_controller_names(self.detector_names)
        ctrl_prefixes = [f"{ctrl_name}:" for ctrl_name in controller_names]
        return any(
            name
            for name in scan.scan_info["channels"].keys()
            if any(name.startswith(ctrl_prefix) for ctrl_prefix in ctrl_prefixes)
        )


def _get_motors_from_expression(
    expression: str,
    start_var: str = "<",
    end_var: str = ">",
) -> Tuple[str, str]:
    pattern = rf"{re.escape(start_var)}([^{re.escape(end_var)}]+){re.escape(end_var)}"
    return re.findall(pattern, expression)


def _get_scan_ranges_from_sequence(scan: BlissScanType):
    final_scan = scan.streams["SUBSCANS"][:][-1]
    return (2, final_scan.info["scan_nb"])



from .. import testing
from ..processors.streamline_id31 import DemoId31StreamlineScanner
from .utils import streamline


def id31_streamline_demo():
    test_streamline_without_autocalibration()
    test_streamline_with_autocalibration()


def id31_streamline_auto_stop_demo():
    test_streamline_id31_auto_stop()


@testing.integration_fixture
def streamline_id31_scanner():
    return DemoId31StreamlineScanner()


@testing.integration_fixture
def streamline_id31_auto_stop_scanner():
    return DemoId31StreamlineScanner()


@testing.integration_test
def test_streamline_with_autocalibration(streamline_id31_scanner):
    streamline.test_streamline_with_autocalibration(streamline_id31_scanner)


@testing.integration_test
def test_streamline_without_autocalibration(streamline_id31_scanner):
    streamline.test_streamline_without_autocalibration(streamline_id31_scanner)


@testing.integration_test
def test_streamline_id31_auto_stop(streamline_id31_auto_stop_scanner):
    streamline.test_streamline_id31_auto_stop(streamline_id31_auto_stop_scanner)



from blissdata.beacon.data import BeaconData

from ..version_utils import has_minimal_version

if has_minimal_version("blissdata", "1.1"):
    from pydantic import Field  # noqa
    from pydantic import field_validator
else:
    from pydantic.v1 import Field  # noqa
    from pydantic.v1 import validator

    def field_validator(*fields, mode="after", **kwargs):
        pre = mode == "before"
        return validator(*fields, pre=pre, **kwargs)


MODEL_ARGUMENTS = {}

if has_minimal_version("redis-om", "1.0.0"):
    MODEL_ARGUMENTS["index"] = True


def get_redis_db_url():
    if has_minimal_version("blissdata", "1.0"):
        return BeaconData().get_redis_db()

    raw_url = BeaconData().get_redis_db()
    _, url = raw_url.split(":")

    if url.endswith("sock"):
        return f"unix://{url}"
    else:
        return f"redis://{url}"



from typing import Any
from typing import Dict
from typing import Optional

from ..fluo.processor import FluoProcessor
from ..processor import BlissScanType


class Id16bFluoProcessor(FluoProcessor):

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = dict()
        defaults.setdefault("instrument_name", "id16b")
        defaults.setdefault("output_suffix", "_xrf")
        super().__init__(config=config, defaults=defaults)

    def _is_xrfmap(self, scan: BlissScanType) -> bool:
        return scan.scan_info["type"] == "fscan2d"

    def _is_fluoxas(self, scan: BlissScanType) -> bool:
        return False

    def _is_mosaic_xrfmap(self, scan: BlissScanType) -> bool:
        return False

    def get_xrfmap_arguments(self, scan: BlissScanType) -> dict:
        kwargs = super().get_xrfmap_arguments(scan)
        kwargs.update(self._get_common_arguments(scan))
        return kwargs

    def get_mosaic_xrfmap_arguments(self, scan: BlissScanType) -> dict:
        kwargs = super().get_mosaic_xrfmap_arguments(scan)
        kwargs.update(self._get_common_arguments(scan))
        return kwargs

    def get_fluoxas_arguments(self, scan: BlissScanType) -> dict:
        kwargs = super().get_fluoxas_arguments(scan)
        kwargs.update(self._get_common_arguments(scan))
        return kwargs

    def _get_common_arguments(self, scan: BlissScanType) -> dict:
        kwargs = {}
        if scan.scan_info["type"] in ("fscan2d", "fscan"):
            kwargs["livetime_ref_value"] = (
                "<instrument/fscan_parameters/fast_step_time>"
            )
        return kwargs



import logging
from functools import wraps
from typing import Callable

from ..import_utils import unavailable_class
from ..version_utils import has_minimal_version

try:
    if has_minimal_version("bliss", "2.2"):
        from flint.client.plots import BasePlot
        from flint.client.plots import ImageView
        from flint.client.plots import Plot1D
        from flint.client.proxy import FlintClient
    else:
        from bliss.flint.client.plots import BasePlot
        from bliss.flint.client.plots import ImageView
        from bliss.flint.client.plots import Plot1D
        from bliss.flint.client.proxy import FlintClient
except ImportError as ex:
    FlintClient = unavailable_class(ex)
    BasePlot = unavailable_class(ex)
    ImageView = unavailable_class(ex)
    Plot1D = unavailable_class(ex)

logger = logging.getLogger(__name__)


def capture_errors(method) -> Callable:
    @wraps(method)
    def wrapper(*args, **kw):
        try:
            return method(*args, **kw)
        except Exception as e:
            msg = f"Flint plot error: {e}"
            logger.error(msg, exc_info=True)

    return wrapper



import logging
from enum import IntEnum
from typing import Tuple

import numpy as np

from blissoda.bliss_globals import current_session
from blissoda.import_utils import unavailable_class
from blissoda.import_utils import unavailable_function
from blissoda.import_utils import unavailable_module

try:
    from bliss.physics import units
    from flint.viewers.custom_image.client import ImageView
    from tomo.globals import get_active_tomo_config
except ImportError as ex:
    units = unavailable_module(ex)
    get_active_tomo_config = unavailable_function(ex)
    ImageView = unavailable_class(ex)


logger = logging.getLogger(__name__)


def get_estimate_cor_metadata() -> tuple:
    """Return (pixel_size_mm, detector_width, translation_y_mm) for CoR estimation."""
    config = current_session.config

    tomo_config = get_active_tomo_config()
    tomo_detector = tomo_config.detectors.active_detector
    lima_detector = tomo_config.detectors.get_detector(
        tomo_config.detectors.active_detector
    )

    pixel_size_mm = tomo_detector.sample_pixel_size / 1000  # default mm

    detector_width = lima_detector.image.fullsize[0]

    translation_y_motor = tomo_config.sample_stage.y_axis.name

    translation_y_motor_config = config.get(translation_y_motor)
    translation_y_value = float(translation_y_motor_config.position)
    translation_y_unit = translation_y_motor_config.unit
    try:  # FIX ME: handle case where unit is not recognized
        translation_y_mm = (
            translation_y_value * units.ur(translation_y_unit).to("mm").magnitude
        )
    except Exception:
        translation_y_mm = translation_y_value

    return pixel_size_mm, detector_width, translation_y_mm


def get_reconstruction_metadata(scan) -> tuple:
    """Return (distance_m, energy_keV, pixel_size_m) from scan metadata."""

    technique_info = scan.scan_info["technique"]
    scan_info = technique_info["scan"]

    # Get distance in meters
    distance = float(scan_info["effective_propagation_distance"])
    distance_unit = scan_info.get("effective_propagation_distance@units", "m")
    distance_m = distance * units.ur(distance_unit).to("m").magnitude

    # Get energy in keV
    energy = float(scan_info["energy"])
    energy_unit = scan_info.get("energy@units", "keV")
    energy_keV = energy * units.ur(energy_unit).to("keV").magnitude

    # Get pixel size in meters
    pixel_size = scan_info["sample_pixel_size"]
    pixel_size_unit = scan_info["sample_pixel_size@units"]
    pixel_size_m = pixel_size * units.ur(pixel_size_unit).to("m").magnitude

    return distance_m, energy_keV, pixel_size_m


def calculate_CoR_estimate(
    pixel_size_mm: float,
    translation_y_mm: float,
    detector_width: int,
    offset_mm: float,
    flip: bool,
    orientation_factor: float = 1.0,
) -> float:
    """
    Estimate the center of rotation for tomographic reconstruction.
    """
    # The -0.5 accounts for pixel-center coordinate convention used by the reconstruction algorithm
    default_center = detector_width / 2 - 0.5

    # This tells us how many pixels the center of rotation has shifted
    pixel_shift = (translation_y_mm - offset_mm) / pixel_size_mm

    # Adjust the center of rotation based on the pixel shift and orientation factor
    center_of_rotation = default_center + (pixel_shift * orientation_factor)

    if flip:
        center_of_rotation *= -1

    return center_of_rotation


def calculate_relative_CoR_estimate(
    pixel_size_mm: float,
    translation_y_mm: float,
    offset_mm: float,
    flip: bool,
    orientation_factor: float = 1.0,
) -> float:
    """
    Estimate the relative center of rotation for tomographic reconstruction.
    """

    # This tells us how many pixels the center of rotation has shifted
    pixel_shift = (translation_y_mm - offset_mm) / pixel_size_mm

    # Adjust the center of rotation based on the pixel shift and orientation factor
    relative_center_of_rotation = pixel_shift * orientation_factor

    if flip:
        relative_center_of_rotation *= -1

    return relative_center_of_rotation


class ImageKey(IntEnum):
    PROJECTION = 0
    DARK_FIELD = 2
    FLAT_FIELD = 1
    INVALID = 3


def compute_axes(image: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute physical axes from sample config; fallback to pixel indices.

    Converts pixel_size (um) to axis units defined by cfg.sample_x_axis.unit
    and cfg.sample_y_axis.unit.

    Parameters
    ----------
    image : np.ndarray
        The image array (for shape information)

    Returns
    -------
    tuple of np.ndarray
        x_axis and y_axis arrays in physical units
    """

    tomo_config = get_active_tomo_config()
    pixel_size_um = tomo_config.detectors.active_detector.sample_pixel_size

    # Determine units for x and y axes
    unit_x = getattr(tomo_config.sample_x_axis, "unit", "mm")
    unit_y = getattr(tomo_config.sample_y_axis, "unit", "mm")

    # Conversion factors from micrometers to axis units
    conv = {"um": 1.0, "mm": 1e-3}
    if unit_x not in conv:
        logger.warning("Unknown unit '%s' for sample_x_axis, assuming 'mm'", unit_x)
    if unit_y not in conv:
        logger.warning("Unknown unit '%s' for sample_y_axis, assuming 'mm'", unit_y)

    # Convert pixel size to axis units
    ps_x = pixel_size_um * conv.get(unit_x, 0.001)
    ps_y = pixel_size_um * conv.get(unit_y, 0.001)

    rows, cols = image.shape[-2:]

    # Center positions are already in axis units
    cx = tomo_config.sample_x_axis.position
    cy = tomo_config.sample_y_axis.position

    half_w = (cols * ps_x) / 2
    half_h = (rows * ps_y) / 2

    x_axis = np.linspace(cx - half_w, cx + half_w, cols)
    y_axis = np.linspace(cy - half_h, cy + half_h, rows)

    return x_axis, y_axis


def apply_labels(widget: ImageView) -> None:
    """
    Apply axis labels from tomo config.

    Parameters
    ----------
    widget : ImageView
        The plot widget
    """
    tomo_config = get_active_tomo_config()
    widget.xlabel = tomo_config.sample_y.name
    widget.ylabel = tomo_config.sample_x.name



import logging
from typing import Any
from typing import Dict
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from .persistent.parameters import WithPersistentParameters
from .utils.classes import NoMethodAssignment

logger = logging.getLogger(__name__)


class BlissAutomationConfig(BaseModel, extra="forbid"):
    """Pydantic model for parsing Beacon configuration of a Bliss object for automation.

    For example:

    .. code-block:: yaml

        - name: my_processor
          plugin: generic
          class: MyProcessor
          package: blissoda.mycategory.my_processor
          default_parameters:
            param1: value1
            param2: value2
          shared: false
          singleton: true

    In this example the Redis key is `blissoda:demo_session:MyProcessor`.
    """

    name: Optional[str] = Field(
        None, description="Name of the instance (used if singleton is True)."
    )
    plugin: Optional[str] = Field(None, description="Plugin type, typically 'generic'.")
    class_: Optional[str] = Field(
        None, alias="class", description="Class name of the object to instantiate."
    )
    package: Optional[str] = Field(
        None, description="Python package where the class is located."
    )
    default_parameters: Dict[str, Any] = Field(
        default_factory=dict,
        description="Initial values for parameters stored in Redis.",
    )
    shared: bool = Field(
        False, description="If True, parameters are shared between Bliss sessions."
    )
    singleton: bool = Field(
        True,
        description="If True, parameters are a singleton for the Bliss session (``shared=False``)"
        "or globally (``shared=True``).",
    )


class BlissAutomationObject(WithPersistentParameters, NoMethodAssignment):
    """A Bliss object for automation (data processing and/or acquisition).

    It has persistent parameters in Redis and configuration in Beacon.

    The Beacon configuration is expected to use the
    `generic plugin <https://bliss.gitlab-pages.esrf.fr/bliss/master/config_plugins.html#plugin-generic>`_.
    """

    CONFIG_MODEL_CLASS = BlissAutomationConfig

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        :param config: Beacon configuration.
        :param defaults: Initial values for parameters stored in Redis;
                         merged with priority over ``config["default_parameters"]``.
        """
        if not issubclass(self.CONFIG_MODEL_CLASS, BlissAutomationConfig):
            raise TypeError(
                "CONFIG_MODEL_CLASS must be a subtype of BlissAutomationConfig"
            )

        config_dict = config or {}
        config_model = self.CONFIG_MODEL_CLASS(**config_dict)

        shared = config_model.shared
        name = None if config_model.singleton else config_model.name

        defaults = defaults or {}
        defaults = {**defaults, **config_model.default_parameters}

        super().__init__(shared=shared, name=name, defaults=defaults)

        self._config = config_model



"""Bliss-side for Flint ID26 plots"""

from ..flint import ImageView
from ..flint import Plot1D


class AggregatedScansPlot(Plot1D):
    """Plot for aggregated scans."""


class RixsPlot(ImageView):
    """Plot for RIXS maps."""



import json
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ...bliss_globals import current_session
from ...bliss_globals import setup_globals
from ...persistent.parameters import ParameterInfo
from ...processor import BlissScanType
from ...utils import directories
from ...xrpd.processor import XrpdProcessor
from .utils.pyfai_calib import DEFAULT_CALIB
from .utils.xrpd import lima_image_is_padded


class DemoXrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_cython",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )

        super().__init__(config=config, defaults=defaults)
        self.queue = "celery"

    def get_integrate_inputs(
        self, scan, lima_name: str, task_identifier: str
    ) -> List[dict]:
        inputs = super().get_integrate_inputs(scan, lima_name, task_identifier)
        inputs.append(
            {
                "task_identifier": task_identifier,
                "name": "demo",
                "value": lima_image_is_padded(lima_name),
            }
        )
        return inputs

    def _get_demo_result_dir(self, dataset_filename: str) -> str:
        root_dir = directories.get_processed_dir(dataset_filename)
        return os.path.join(root_dir, "demo", "xrpd")

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_demo_result_dir(dataset_filename)
        return os.path.join(root_dir, "workflows")

    def _get_config_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_demo_result_dir(dataset_filename)
        return os.path.join(root_dir, "config")

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        root_dir = self._get_config_dir(current_session.scan_saving.filename)
        config_filename = os.path.join(root_dir, f"pyfaicalib_{lima_name}.json")

        if not os.path.exists(config_filename):
            os.makedirs(os.path.dirname(config_filename), exist_ok=True)

            poni = DEFAULT_CALIB[lima_name]
            if "detector_config" in poni:
                self.integration_options["version"] = 5

            with open(config_filename, "w") as f:
                json.dump(poni, f)

        return config_filename

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        return self.integration_options.to_dict()

    def pct(self, *args, **kw):
        s = setup_globals.ct(*args, **kw)
        self.on_new_scan(s)
        return s

    def _expected_data_shapes_1d(self, lima_name: str, npoints: int):
        return {
            f"{lima_name}:q": (4096,),
            f"{lima_name}:intensity": (npoints, 4096),
            f"{lima_name}:points": (npoints,),
        }



import os
from pprint import pprint
from typing import Any
from typing import Dict
from typing import Optional

from ...bliss_globals import current_session
from ...id22.stscan_processor import StScanProcessor
from ...utils import directories


class DemoStScanProcessor(StScanProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        if self._HAS_BLISS:
            root_dir = directories.get_processed_dir(
                current_session.scan_saving.filename
            )
            root_dir = os.path.join(root_dir, "demo", "id22", "config")

            defaults.setdefault(
                "_convert_workflow",
                os.path.join(root_dir, "convert.json"),
            )
            defaults.setdefault(
                "_rebinsum_workflow",
                os.path.join(root_dir, "rebinsum.json"),
            )
            defaults.setdefault(
                "_extract_workflow",
                os.path.join(root_dir, "extract.json"),
            )

        super().__init__(config=config, defaults=defaults)

    def _submit_job(self, workflow, inputs, convert_destination, **kw):
        print("\nSubmit workfow")
        print(workflow)
        print("Inputs:")
        pprint(inputs)
        print("Save for provenance:")
        pprint(convert_destination)
        print("Options:")
        pprint(kw)

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        return os.path.join(directories.get_workflows_dir(dataset_filename), "id22")



import logging
import warnings
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

from . import defaults

logger = logging.getLogger(__name__)


def get_detector_names(
    detectors: Optional[Sequence[Union[int, str]]] = None,
    detector_numbers: Optional[Sequence[int]] = None,
    detector_names: Optional[Sequence[str]] = None,
    instrument_name: Optional[str] = None,
) -> List[str]:
    # Check mutually exclusive arguments
    provided = [
        detectors is not None,
        detector_numbers is not None,
        detector_names is not None,
    ]
    if sum(provided) > 1:
        raise ValueError(
            "Arguments 'detectors', 'detector_numbers', and 'detector_names' are mutually exclusive."
        )

    # Handle deprecated arguments
    if detector_numbers is not None:
        msg = "'detector_numbers' is deprecated; use 'detectors' instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        logger.warning(msg)
        detectors = detector_numbers

    if detector_names is not None:
        msg = "'detector_names' is deprecated; use 'detectors' instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        logger.warning(msg)
        detectors = detector_names

    name_format = defaults.MCA_NAME_FORMAT.get(instrument_name)
    names: List[str] = []
    for det in detectors:
        if isinstance(det, int):
            if not name_format:
                raise ValueError(
                    f"No MCA name format is known for instrument {instrument_name!r}"
                )
            names.append(name_format.format(det))
        elif isinstance(det, str):
            names.append(det)
        else:
            raise TypeError(f"Detector must be int or str, got {type(det)}")

    return names



from .. import testing
from ..processors.id22 import DemoStScanProcessor


@testing.integration_fixture
def stscan_processor():
    yield DemoStScanProcessor()


@testing.integration_test
def id22_stscan_demo(stscan_processor):
    stscan_processor.submit_workflows()
